"""Formula 7.9 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form7Dot9EpsilonSmMinusEpsilonCm(Formula):
    r"""Class representing formula 7.9 for the calculation of [$\epsilon_{sm} - \epsilon_{cm}$]."""

    label = "7.9"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        sigma_s: MPA,
        k_t: DIMENSIONLESS,
        f_ct_eff: MPA,
        rho_p_eff: DIMENSIONLESS,
        e_s: MPA,
        e_cm: MPA,
    ) -> None:
        r"""[$\epsilon_{sm} - \epsilon_{cm}$] Calculation of the strain difference [$\epsilon$].

        EN 1992-1-1:2004 art.7.3.4(2) - Formula (7.9)

        Parameters
        ----------
        sigma_s : MPA
            [$\sigma_s$] Stress in the reinforcement [$MPa$].
        k_t : DIMENSIONLESS
            [$k_t$] Factor dependent on the duration of the load, 0.6 for short term loading, 0.4 for long term loading [$-$].
        f_ct_eff : MPA
            [$f_{ct,eff}$] Effective tensile strength of concrete [$MPa$].
        rho_p_eff : DIMENSIONLESS
            [$\rho_{p,eff}$] Effective reinforcement ratio, see equation 7.10 [$-$].
        e_s : MPA
            [$e_s$] Modulus of elasticity of reinforcement [$MPa$].
        e_cm : MPA
            [$e_{cm}$] Modulus of elasticity of concrete [$MPa$].
        """
        super().__init__()
        self.sigma_s = sigma_s
        self.k_t = k_t
        self.f_ct_eff = f_ct_eff
        self.rho_p_eff = rho_p_eff
        self.e_s = e_s
        self.e_cm = e_cm
        self.alpha_e = e_s / e_cm

    @staticmethod
    def _evaluate(
        sigma_s: MPA,
        k_t: DIMENSIONLESS,
        f_ct_eff: MPA,
        rho_p_eff: DIMENSIONLESS,
        e_s: MPA,
        e_cm: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(sigma_s=sigma_s, k_t=k_t, f_ct_eff=f_ct_eff, rho_p_eff=rho_p_eff, e_s=e_s, e_cm=e_cm)

        alpha_e = e_s / e_cm
        numerator = sigma_s - k_t * (f_ct_eff / rho_p_eff) * (1 + alpha_e * rho_p_eff)
        return max(numerator / e_s, 0.6 * sigma_s / e_s)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.9."""
        _equation: str = (
            r"\max\left("
            r"\frac{\sigma_s - k_t \cdot \frac{f_{ct,eff}}{\rho_{p,eff}} \cdot \left(1 + \frac{E_s}{E_{cm}} \cdot \rho_{p,eff}\right)}{E_s}; "
            r"\frac{0.6 \cdot \sigma_s}{E_s}"
            r"\right)"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\sigma_s": f"{self.sigma_s:.{n}f}",
                r"k_t": f"{self.k_t:.{n}f}",
                r"f_{ct,eff}": f"{self.f_ct_eff:.{n}f}",
                r"\rho_{p,eff}": f"{self.rho_p_eff:.{n}f}",
                r"E_s": f"{self.e_s:.{n}f}",
                r"E_{cm}": f"{self.e_cm:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\epsilon_{sm} - \epsilon_{cm}",
            result=f"{self:.6f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 7.6 from EN 1995-1-1:2004."""

from contextlib import AbstractContextManager
from contextlib import nullcontext as does_not_raise

import pytest

from blueprints.codes.eurocode.en_1995_1_1_2004.chapter_7_serviceability_limit_states.formula_7_6 import Form7Dot6VelocityResponse
from blueprints.validations import LessOrEqualToZeroError


class TestForm7Dot6VelocityResponse:
    """Validation for formula 7.6 from EN 1995-1-1:2004."""

    @pytest.mark.parametrize(
        ("n_40", "m", "length", "b", "expected_result"),
        [(0.77, 226, 5.30, 1.00, 0.002466), (1, 1, 1, 1, 0.0199)],
        ids=["test 1", "test 2"],
    )
    def test_evaluation(self, n_40: float, m: float, length: float, b: float, expected_result: float) -> None:
        """Test the evaluation of the result."""
        form_7_6 = Form7Dot6VelocityResponse(n_40=n_40, m=m, length=length, b=b)
        assert form_7_6 == pytest.approx(expected=expected_result, rel=1e-3)

    @pytest.mark.parametrize(
        ("n_40", "m", "length", "b", "expectation"),
        [
            (1, 1, 1, 1, does_not_raise()),
            (0.01, 0.2, 10, 2, does_not_raise()),
            (0, 1, 1, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, 0, 1, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, 1, 0, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, 1, 1, 0, pytest.raises(LessOrEqualToZeroError)),
            (-1, 1, 1, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, -1, 1, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, 1, -1, 1, pytest.raises(LessOrEqualToZeroError)),
            (1, 1, 1, -1, pytest.raises(LessOrEqualToZeroError)),
        ],
        ids=[
            "passes test 1",
            "passes test 2",
            "zero_n_40",
            "zero_m",
            "zero_length",
            "zero_b",
            "negative_n_40",
            "negative_m",
            "negative_length",
            "negative_b",
        ],
    )
    def test_raise_error_incorrect_args(self, n_40: float, m: float, length: float, b: float, expectation: AbstractContextManager) -> None:
        """Test the evaluation of the result."""
        with expectation:
            assert Form7Dot6VelocityResponse(n_40=n_40, length=length, m=m, b=b) is not None

    @pytest.mark.parametrize(
        ("n_40", "m", "length", "b", "rep_short", "rep_long"),
        [
            (
                1,  # HZ
                1,  # KG_M2
                1,  # M
                1,  # M
                r"v = 0.020",
                r"v = \frac{4 \cdot (0.4 + 0.6 \cdot n_{40})}{m \cdot b \cdot l + 200} = \frac{4 \cdot (0.4 + 0.6 \cdot 1.00)}{1.00 \cdot 1.00 \cdot 1.00 + 200} = 0.020",  # noqa: E501
            ),
            (
                0.77,  # HZ
                226,  # KG_M2
                5.3,  # M
                1,  # M
                r"v = 0.002",
                r"v = \frac{4 \cdot (0.4 + 0.6 \cdot n_{40})}{m \cdot b \cdot l + 200} = \frac{4 \cdot (0.4 + 0.6 \cdot 0.77)}{226.00 \cdot 1.00 \cdot 5.30 + 200} = 0.002",  # noqa: E501
            ),
        ],
        ids=["test with integer input", "test with decimal inputs"],
    )
    def test_latex(self, n_40: float, m: float, length: float, b: float, rep_short: str, rep_long: str) -> None:
        """Test the latex representation of the formula."""
        # act
        form_7_6_latex = Form7Dot6VelocityResponse(n_40=n_40, b=b, m=m, length=length).latex()

        # assert
        assert form_7_6_latex.complete == rep_long
        assert form_7_6_latex.short == rep_short



"""Testing formula 5.10b from EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_10b import Form5Dot10bRedistributionOfMomentsUpperFck
from blueprints.validations import LessOrEqualToZeroError


class TestForm5Dot10bRedistributionOfMomentsUpperFck:
    """Validation for formula 5.10b from EN 1992-1-1:2004."""

    @pytest.mark.parametrize(
        ("delta", "k3", "k4", "xu", "d", "result_manual"),
        [
            (0.5, 0.44, 1.25, 0.2, 0.5, False),
            (1.0, 0.44, 1.25, 0.2, 0.5, True),
        ],
    )
    def test_comparison(self, delta: float, k3: float, k4: float, xu: float, d: float, result_manual: bool) -> None:
        """Test the evaluation of the comparison."""
        form_5_10b = bool(Form5Dot10bRedistributionOfMomentsUpperFck(delta=delta, k3=k3, k4=k4, xu=xu, d=d))
        assert form_5_10b == pytest.approx(expected=result_manual)

    @pytest.mark.parametrize(
        ("delta", "k3", "k4", "xu", "d"),
        [
            (0.5, 0.44, 1.25, 0.2, 0.5),
            (1.0, 0.44, 1.25, 0.2, 0.5),
        ],
    )
    def test_evaluation(self, delta: float, k3: float, k4: float, xu: float, d: float) -> None:
        """Test the right-hand side evaluation of the result."""
        form_5_10b = Form5Dot10bRedistributionOfMomentsUpperFck(delta=delta, k3=k3, k4=k4, xu=xu, d=d)
        expected_rhs = k3 + k4 * (xu / d)
        assert form_5_10b.right_hand_side == pytest.approx(expected_rhs, abs=1e-3)

    def test_raise_error_if_negative_values(self) -> None:
        """Test that a LessOrEqualToZeroError is raised when a negative or zero value is passed."""
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot10bRedistributionOfMomentsUpperFck(delta=-0.5, k3=0.44, k4=1.25, xu=0.2, d=0.5)
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot10bRedistributionOfMomentsUpperFck(delta=0.5, k3=0, k4=1.25, xu=0.2, d=0.5)

    @pytest.mark.parametrize(
        ("delta", "k3", "k4", "xu", "d", "expected_lhs_rhs"),
        [
            (0.94, 0.44, 1.25, 0.2, 0.5, 0.94),
            (0.825, 0.2, 1.25, 0.3, 0.6, 0.825),
        ],
    )
    def test_properties(self, delta: float, k3: float, k4: float, xu: float, d: float, expected_lhs_rhs: float) -> None:
        """Test the properties of the comparison."""
        form_5_10b = Form5Dot10bRedistributionOfMomentsUpperFck(delta=delta, k3=k3, k4=k4, xu=xu, d=d)
        assert form_5_10b.left_hand_side == pytest.approx(expected_lhs_rhs, abs=1e-3)
        assert form_5_10b.right_hand_side == pytest.approx(expected_lhs_rhs, abs=1e-3)
        assert form_5_10b.ratio == pytest.approx(1, abs=1e-3)

    def test_str_representation(self) -> None:
        """Test the string representation of the comparison."""
        form_5_10b = Form5Dot10bRedistributionOfMomentsUpperFck(delta=0.5, k3=0.44, k4=1.25, xu=0.2, d=0.5)
        expected_str = (
            r"CHECK \rightarrow \delta \geq k_3 + k_4 \frac{x_u}{d} \rightarrow 0.500 \geq 0.44 + 1.25 \frac{0.2}{" r"0.5} \rightarrow \text{Not OK}"
        )
        assert str(form_5_10b) == expected_str

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow \delta \geq k_3 + k_4 \frac{x_u}{d} \rightarrow 0.500 \geq 0.44 + 1.25 \frac{"
                r"0.2}{0.5} \rightarrow \text{Not OK}",
            ),
            ("short", r"CHECK \rightarrow \text{Not OK}"),
        ],
    )
    def test_latex_ok(self, representation: str, expected: str) -> None:
        """Test the latex representation of the comparison."""
        # Example values
        delta = 0.5
        k3 = 0.44
        k4 = 1.25
        xu = 0.2
        d = 0.5

        # Object to test
        form_5_10b_latex = Form5Dot10bRedistributionOfMomentsUpperFck(delta=delta, k3=k3, k4=k4, xu=xu, d=d).latex()

        actual = {
            "complete": form_5_10b_latex.complete,
            "short": form_5_10b_latex.short,
        }

        assert actual[representation] == expected

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \rightarrow \delta \geq k_3 + k_4 \frac{x_u}{d} \rightarrow 2000.000 \geq 0.44 + 1.25 \frac{" r"200}{50} \rightarrow OK",
            ),
            ("short", r"CHECK \rightarrow OK"),
        ],
    )
    def test_latex_not_ok(self, representation: str, expected: str) -> None:
        """Test the latex representation of the comparison."""
        # Example values
        delta = 2000
        k3 = 0.44
        k4 = 1.25
        xu = 200
        d = 50

        # Object to test
        form_5_10b_latex = Form5Dot10bRedistributionOfMomentsUpperFck(delta=delta, k3=k3, k4=k4, xu=xu, d=d).latex()

        actual = {
            "complete": form_5_10b_latex.complete,
            "short": form_5_10b_latex.short,
        }

        assert actual[representation] == expected



"""Formula 3.5 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot5ApproximationVarianceElasticModulusOverTime(Formula):
    """Class representing formula 3.5 for the approximation of the elastic modulus, Ecm(t) at day t."""

    label = "3.5"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_cm_t: MPA,
        f_cm: MPA,
        e_cm: MPA,
    ) -> None:
        r"""[$E_{cm}(t)$] The approximated elastic modulus at day t [$MPa$].

        EN 1992-1-1:2004 art.3.1.3(3) - Formula (3.5)

        Parameters
        ----------
        f_cm_t : MPA
            [$f_{cm}(t)$] Compressive strength concrete at t days [$MPa$].
        f_cm : MPA
            [$f_{cm}$] Average concrete compressive strength on day 28 based on table 3.1 [$MPa$].
        e_cm : MPA
            [$E_{cm}$] Average elastic modulus on day 28 [$MPa$].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_cm_t = f_cm_t
        self.f_cm = f_cm
        self.e_cm = e_cm

    @staticmethod
    def _evaluate(
        f_cm_t: MPA,
        f_cm: MPA,
        e_cm: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_cm_t < 0:
            raise ValueError(f"Negative f_cm_t: {f_cm_t}. f_cm_t cannot be negative")
        if f_cm < 0:
            raise ValueError(f"Negative f_cm: {f_cm}. f_cm cannot be negative")
        if e_cm < 0:
            raise ValueError(f"Negative e_cm: {e_cm}. e_cm cannot be negative")
        return (f_cm_t / f_cm) ** 0.3 * e_cm

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.5."""
        return LatexFormula(
            return_symbol=r"E_{cm}(t)",
            result=f"{self:.{n}f}",
            equation=r"( f_{cm}(t) / f_{cm} )^{0.3} \cdot E_{cm}",
            numeric_equation=rf"( {self.f_cm_t:.{n}f} / {self.f_cm:.{n}f} )^{{0.3}} \cdot {self.e_cm:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 3.17 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_17 import Form3Dot17CompressiveStressConcrete


class TestForm3Dot17CompressiveStressConcrete:
    """Validation for formula 3.17 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        f_cd = 18.50  # MPa
        epsilon_c = 0.64  # -
        epsilon_c2 = 0.81  # -
        n = 2.0  # -

        form_3_17 = Form3Dot17CompressiveStressConcrete(f_cd=f_cd, epsilon_c=epsilon_c, epsilon_c2=epsilon_c2, n=n)

        # Expected result, manually calculated
        manually_calculated_result = 17.68511

        assert form_3_17 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_f_cd_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        f_cd = -18.50  # MPa
        epsilon_c = 0.64  # -
        epsilon_c2 = 0.81  # -
        n = 2.0  # -

        with pytest.raises(ValueError):
            Form3Dot17CompressiveStressConcrete(f_cd=f_cd, epsilon_c=epsilon_c, epsilon_c2=epsilon_c2, n=n)

    def test_raise_error_when_negative_epsilon_c_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        f_cd = 18.50  # MPa
        epsilon_c = -0.64  # -
        epsilon_c2 = 0.81  # -
        n = 2.0  # -

        with pytest.raises(ValueError):
            Form3Dot17CompressiveStressConcrete(f_cd=f_cd, epsilon_c=epsilon_c, epsilon_c2=epsilon_c2, n=n)

    def test_raise_error_when_epsilon_c_is_larger_than_epsilon_c2(self) -> None:
        """Test a negative value."""
        # Example values
        f_cd = 18.50  # MPa
        epsilon_c = 1.64  # -
        epsilon_c2 = 0.81  # -
        n = 2.0  # -

        with pytest.raises(ValueError):
            Form3Dot17CompressiveStressConcrete(f_cd=f_cd, epsilon_c=epsilon_c, epsilon_c2=epsilon_c2, n=n)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\sigma_c = f_{cd} \cdot \left[ 1 - \left( 1 - \frac{\epsilon_c}{\epsilon_{c2}} \right)^n \right] = "
                r"18.500 \cdot \left[ 1 - \left( 1 - \frac{0.640}{0.810} \right)^{2.000} \right] = 17.685",
            ),
            ("short", r"\sigma_c = 17.685"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        f_cd = 18.50  # MPa
        epsilon_c = 0.64  # -
        epsilon_c2 = 0.81  # -
        n = 2.0  # -

        # Object to test
        form_3_17_latex = Form3Dot17CompressiveStressConcrete(f_cd=f_cd, epsilon_c=epsilon_c, epsilon_c2=epsilon_c2, n=n).latex()

        actual = {"complete": form_3_17_latex.complete, "short": form_3_17_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.22 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_22 import Form6Dot22CheckCrushingCompressionStruts
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot22CheckCrushingCompressionStruts:
    """Validation for formula 6.22 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 8.0  # MPA
        nu = 0.6  # dimensionless
        f_cd = 30.0  # MPA
        theta_f = 45.0  # DEG

        # Object to test
        formula = Form6Dot22CheckCrushingCompressionStruts(v_ed=v_ed, nu=nu, f_cd=f_cd, theta_f=theta_f)

        # Expected result, manually calculated
        expected_result = True

        assert formula == expected_result

    @pytest.mark.parametrize(
        ("v_ed", "nu", "f_cd", "theta_f"),
        [
            (-8.0, 0.6, 30.0, 45.0),  # v_ed is negative
            (8.0, -0.6, 30.0, 45.0),  # nu is negative
            (8.0, 0.6, -30.0, 45.0),  # f_cd is negative
            (8.0, 0.6, 30.0, -45.0),  # theta_f is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, nu: float, f_cd: float, theta_f: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot22CheckCrushingCompressionStruts(v_ed=v_ed, nu=nu, f_cd=f_cd, theta_f=theta_f)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"CHECK \to v_{Ed} \leq \nu \cdot f_{cd} \cdot \sin(\theta_{f}) \cdot \cos(\theta_{f}) \to "
                r"8.000 \leq 0.600 \cdot 30.000 \cdot \sin(45.000) \cdot \cos(45.000) \to OK",
            ),
            ("short", r"CHECK \to OK"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 8.0  # MPa
        nu = 0.6  # dimensionless
        f_cd = 30.0  # MPA
        theta_f = 45.0  # DEG

        # Object to test

        latex = Form6Dot22CheckCrushingCompressionStruts(v_ed=v_ed, nu=nu, f_cd=f_cd, theta_f=theta_f).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Test the GU enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.usections import USections


class TestUProfiles:
    """Tests for the GU enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert USections.GU_6N.value == (
            "GU 6N",
            600,
            309,
            6.0,
            6.0,
            248,
            42.5,
            89.0,
            41.9,
            9670,
            625,
            375,
            765,
            69.9,
            10.43,
            1.26,
            "ArcelorMittal",
        )
        assert USections.PU_32_PLUS_1.value == (
            "PU 32+1",
            600,
            452,
            20.5,
            11.4,
            342,
            68.1,
            251.3,
            118.4,
            75410,
            3340,
            1905,
            3845,
            197.3,
            17.32,
            1.52,
            "ArcelorMittal",
        )

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "GU 6N" in [e.value[0] for e in USections]
        assert "PU 12" in [e.value[0] for e in USections]
        assert "PU 32+1" in [e.value[0] for e in USections]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in USections]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = USections.GU_6N
        assert profile.alias == "GU 6N"
        assert profile.b_width_single_pile == 600
        assert profile.h_height_pile == 309
        assert profile.tf_flange_thickness == 6.0
        assert profile.tw_web_thickness == 6.0
        assert profile.bf_flange_width == 248
        assert profile.a_flange_angle == 42.5
        assert profile.a_cross_sectional_area == 89.0
        assert profile.gsp_mass_per_single_pile == 41.9
        assert profile.i_y_moment_inertia == 9670
        assert profile.w_el_y_elastic_section_modulus == 625
        assert profile.s_y_static_moment == 375
        assert profile.w_pl_y_plastic_section_modulus == 765
        assert profile.gw_mass_per_m == 69.9
        assert profile.radius_of_gyration_y_y == 10.43
        assert profile.al_coating_area == 1.26
        assert profile.manufacturer == "ArcelorMittal"
        assert profile.sheet_pile_type == "U-Section"

    def test_profile_count(self) -> None:
        """Test that all profiles are present."""
        # There should be 40 GU/PU profiles
        assert len(USections) == 40

    def test_specific_profiles(self) -> None:
        """Test specific profiles from the JSON data."""
        # Test GU 33N
        profile = USections.GU_33N
        assert profile.alias == "GU 33N"
        assert profile.h_height_pile == 452
        assert profile.tf_flange_thickness == 20.5
        assert profile.tw_web_thickness == 11.4

        # Test PU 12
        profile = USections.PU_12
        assert profile.alias == "PU 12"
        assert profile.h_height_pile == 360
        assert profile.tf_flange_thickness == 9.8
        assert profile.tw_web_thickness == 9.0

        # Test GU 16-400
        profile = USections.GU_16_400
        assert profile.alias == "GU 16-400"
        assert profile.b_width_single_pile == 400
        assert profile.h_height_pile == 290
        assert profile.a_flange_angle == 82.1



"""Formula 5.21 from EN 1992-1-1:2004: Chapter 5 - Structural Analysis."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM4, MPA
from blueprints.validations import raise_if_negative


class Form5Dot21NominalStiffness(Formula):
    r"""Class representing formula 5.21 for the calculation of the nominal stiffness of slender compression members
    with arbitrary cross-section, [$EI$] [$Nmm^2$].
    """

    label = "5.21"
    source_document = EN_1992_1_1_2004

    def __init__(self, k_c: DIMENSIONLESS, e_cd: MPA, i_c: MM4, k_s: DIMENSIONLESS, e_s: MPA, i_s: MM4) -> None:
        r"""[$EI$] Nominal stiffness of slender compression members with arbitrary cross-section.

        EN 1992-1-1:2004 art.5.8.7.2(2) or (3) - Formula (5.21)

        Parameters
        ----------
        k_c : DIMENSIONLESS
            [$K_{c}$] is a factor for effects of cracking, creep etc. see 5.8.7.2 (2) or (3).
        e_cd : MPA
            [$E_{cd}$] is the design value of the modulus of elasticity of concrete, see 5.8.6 (3)
        i_c : MPA
            [$I_{c}$] is the moment of inertia of concrete cross section.
        k_s : DIMENSIONLESS
            [$K_{s}$] is a factor for contribution of reinforcement, see 5.8.7.2 (2) or (3).
        e_s : MPA
            [$E_{s}$] is the design value of the modulus of elasticity of reinforcement, 5.8.6 (3).
        i_s : MPA
            [$I_{s}$] is the second moment of area of reinforcement, about the centre of area of the concrete.
        """
        super().__init__()
        self.k_c = k_c
        self.e_cd = e_cd
        self.i_c = i_c
        self.k_s = k_s
        self.e_s = e_s
        self.i_s = i_s

    @staticmethod
    def _evaluate(k_c: DIMENSIONLESS, e_cd: MPA, i_c: MPA, k_s: DIMENSIONLESS, e_s: MPA, i_s: MPA) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(k_c=k_c, e_cd=e_cd, i_c=i_c, k_s=k_s, e_s=e_s, i_s=i_s)
        return k_c * e_cd * i_c + k_s * e_s * i_s

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 5.21."""
        return LatexFormula(
            return_symbol=r"EI",
            result=f"{self:.{n}f}",
            equation=r"K_{c} \cdot E_{cd} \cdot I_{c} + K_{s} \cdot E_{s} \cdot I_{s}",
            numeric_equation=rf"{self.k_c:.{n}f} \cdot {self.e_cd:.{n}f} "
            rf"\cdot {self.i_c:.{n}f} + {self.k_s:.{n}f} \cdot {self.e_s:.{n}f} \cdot {self.i_s:.{n}f}",
            comparison_operator_label="=",
        )



"""Testing formula 8.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_8_detailing_of_reinforcement_and_prestressing_tendons.formula_8_2 import (
    Form8Dot2UltimateBondStress,
    SubForm8Dot2CoefficientBarDiameter,
    SubForm8Dot2CoefficientQualityOfBond,
)
from blueprints.validations import NegativeValueError


class TestForm8Dot2UltimateBondStress:
    """Validation for formula 8.2 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        eta_1 = 1  # -
        eta_2 = 1  # −
        f_ctd = 20  # MPa
        form_8_2 = Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

        # Expected result, manually calculated
        manually_calculated_result = 45  # MPa

        assert form_8_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_eta_1_is_given(self) -> None:
        """Test if NegativeValueError is raised when a negative value is given for eta_1."""
        # Example values
        eta_1 = -1  # -
        eta_2 = 1  # −
        f_ctd = 20  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

    def test_raise_error_when_negative_eta_2_is_given(self) -> None:
        """Test if NegativeValueError is raised when a negative value is given for eta_2."""
        # Example values
        eta_1 = 1  # -
        eta_2 = -1  # −
        f_ctd = 20  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

    def test_raise_error_when_negative_f_ctd_is_given(self) -> None:
        """Test if NegativeValueError is raised when a negative value is given for f_ctd."""
        # Example values
        eta_1 = 1  # -
        eta_2 = 1  # −
        f_ctd = -20  # MPa

        with pytest.raises(NegativeValueError):
            Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

    def test_integration_with_sub_formula_8_2_coefficient_quality_of_bond(self) -> None:
        """Test the integration with sub-formula 8.2 for the coefficient for quality of bond η1."""
        # Example values
        eta_1 = SubForm8Dot2CoefficientQualityOfBond(bond_quality="good")
        eta_2 = 1  # −
        f_ctd = 20  # MPa

        # Object to test
        form_8_2 = Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

        # Expected result, manually calculated
        manually_calculated_result = 45

        assert form_8_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_integration_with_sub_formula_8_2_coefficient_bar_diameter(self) -> None:
        """Test the integration with sub-formula 8.2 for the coefficient for bar diameter η2."""
        # Example values
        eta_1 = 1  # −
        eta_2 = SubForm8Dot2CoefficientBarDiameter(16)
        f_ctd = 20  # MPa

        # Object to test
        form_8_2 = Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd)

        # Expected result, manually calculated
        manually_calculated_result = 45

        assert form_8_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("representation", "expected_result"),
        [
            (
                "complete",
                r"f_{bd} = 2.25 \cdot \eta_1 \cdot \eta_2 \cdot f_{ctd} = 2.25 \cdot 1.00 \cdot 1.00 \cdot 20.00 = 45.00",
            ),
            ("short", "f_{bd} = 45.00"),
        ],
    )
    def test_latex(self, representation: str, expected_result: str) -> None:
        """Test the LaTeX representation."""
        eta_1 = 1  # -
        eta_2 = 1  # −
        f_ctd = 20  # MPa

        latex = Form8Dot2UltimateBondStress(eta_1=eta_1, eta_2=eta_2, f_ctd=f_ctd).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert actual[representation] == expected_result, f"{representation} representation failed."



"""Test the IPE enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.ipe import IPE


class TestIPE:
    """Tests for the IPE enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert IPE.IPE80.value == ("IPE80", 80, 46, 3.8, 5.2, 5)
        assert IPE.IPE600.value == ("IPE600", 600, 220, 12.0, 19.0, 24)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "IPE80" in [e.value[0] for e in IPE]
        assert "IPE600" in [e.value[0] for e in IPE]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in IPE]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = IPE.IPE80
        assert profile.alias == "IPE80"
        assert profile.total_height == 80
        assert profile.top_flange_width == 46
        assert profile.top_flange_thickness == 5.2
        assert profile.bottom_flange_width == 46
        assert profile.bottom_flange_thickness == 5.2
        assert profile.web_thickness == 3.8
        assert profile.top_radius == 5
        assert profile.bottom_radius == 5



"""Formula 8.7 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_max_curly_brackets
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form8Dot7MinimumCompressionAnchorage(Formula):
    """Class representing formula 8.7 for the calculation of the minimum anchorage length if no other limitation is applied for anchorage in
    compression.
    """

    label = "8.7"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        l_b_rqd: MM,
        diameter: MM,
    ) -> None:
        r"""[$l_{b,min}$] Minimum anchorage length if no other limitation is applied for anchorage in compression. [$mm$].

        EN 1992-1-1:2004 art.8.4.4(1) - Formula (8.7)

        Parameters
        ----------
        l_b_rqd: MM
            [$l_{b,rqd}$] Basic required anchorage length, for anchoring the force [$A_s \cdot \sigma_{sd}$] in a straight bar assuming constant
            bond stress (formula 8.3) [$mm$].
            Use your own implementation for this value or use the Form8Dot3RequiredAnchorageLength class.
        diameter: MM
            [$Ø$] Diameter of the bar [$mm$].
        """
        super().__init__()
        self.l_b_rqd = l_b_rqd
        self.diameter = diameter

    @staticmethod
    def _evaluate(l_b_rqd: MM, diameter: MM) -> MM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(diameter=diameter, l_b_rqd=l_b_rqd)
        return max(0.6 * l_b_rqd, 10 * diameter, 100)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns a LatexFormula object for this formula."""
        return LatexFormula(
            return_symbol=r"l_{b,min}",
            result=f"{self:.{n}f}",
            equation=latex_max_curly_brackets(r"0.6 \cdot l_{b,rqd}", r"10 \cdot Ø", r"100 \ \text{mm}"),
            numeric_equation=latex_max_curly_brackets(
                rf"0.6 \cdot {self.l_b_rqd:.{n}f}",
                rf"10 \cdot {self.diameter}",
                r"100",
            ),
            comparison_operator_label="=",
        )



"""Geometry operations module."""

import math
from enum import Enum, auto

import numpy as np
from shapely import LinearRing, Point

from blueprints.type_alias import RAD


class CoordinateSystemOptions(Enum):
    """Enum of the coordinate system options."""

    XY = auto()
    XZ = auto()
    YZ = auto()


def calculate_rotation_angle(
    start_point: Point,
    end_point: Point,
    coordinate_system: CoordinateSystemOptions = CoordinateSystemOptions.XY,
) -> RAD:
    """
    Calculates rotation of an end point in relation to the start point in a given plane/coordinate system [rad].

    - Start point lies in the origin center of the quadrant.
    - The rotation is calculated in the given plane in relation to the start point.
    - The rotation is calculated in the range [0, 2*pi].
    - The rotation is calculated in the counter-clockwise direction.

    Parameters
    ----------
    start_point: Point
        Starting point that will be used as reference.
    end_point: Point
        End point
    coordinate_system: CoordinateSystemOptions
        Desired plane that will be used as reference to calculate the rotation. Standard is XY-plane.

    Returns
    -------
    RAD
        rotation of end point relative to the start point in radians in the given plane [rad].

    Raises
    ------
    ValueError
        If start_point and end_point are the same.
        If start_point or end_point do not have z value when rotation angle in XZ or YZ plane is requested.
    """
    if list(start_point.coords) == list(end_point.coords):
        msg = f"Start and end point can't be equal. {start_point=} | {end_point=}"
        raise ValueError(msg)

    if coordinate_system != CoordinateSystemOptions.XY and (not start_point.has_z or not end_point.has_z):
        msg = f"Coordinate system '{coordinate_system}' requires z value in both points."
        raise ValueError(msg)

    # Map coordinate system to the corresponding attributes
    coordinates_map = {
        CoordinateSystemOptions.XY: ("x", "y"),
        CoordinateSystemOptions.XZ: ("x", "z"),
        CoordinateSystemOptions.YZ: ("y", "z"),
    }
    horizontal_attr, vertical_attr = coordinates_map[coordinate_system]

    # Extract the coordinates for the specified plane
    horizontal_1, vertical_1 = getattr(start_point, horizontal_attr), getattr(start_point, vertical_attr)
    horizontal_2, vertical_2 = getattr(end_point, horizontal_attr), getattr(end_point, vertical_attr)

    # Calculate the differences in coordinates
    dx = horizontal_2 - horizontal_1
    dy = vertical_2 - vertical_1

    # Calculate the angle using atan2 for correct quadrant determination
    angle = math.atan2(dy, dx)

    # Normalize angle to be in the range [0, 2*pi]
    if angle < 0:
        angle += 2 * math.pi

    return angle


def rotate_linearring(linearring: LinearRing, angle_degrees: float) -> LinearRing:
    """
    Rotate a Shapely LinearRing around its centroid by a specified angle.

    Parameters
    ----------
    linearring : LinearRing
        The Shapely LinearRing to be rotated
    angle_degrees : float
        The rotation angle in degrees (positive for counterclockwise)

    Returns
    -------
    LinearRing
        A new LinearRing that has been rotated

    Examples
    --------
    >>> from shapely.geometry import LinearRing
    >>> square = LinearRing([(0, 0), (1, 0), (1, 1), (0, 1)])
    >>> rotated = rotate_linearring(square, 45)
    """
    # Convert angle from degrees to radians
    angle_radians = np.radians(angle_degrees)

    # Get the centroid as the rotation point
    centroid = linearring.centroid

    # Extract coordinates from the LinearRing
    coords = list(linearring.coords)

    # Create rotation matrix
    rotation_matrix = np.array([[np.cos(angle_radians), -np.sin(angle_radians)], [np.sin(angle_radians), np.cos(angle_radians)]])

    # Initialize an empty list for the rotated coordinates
    rotated_coords = []

    # Define a threshold for floating-point precision
    presition_threshold = 1e-10

    # Rotate each point around the centroid
    for point in coords:
        # Translate point to origin (subtract centroid coordinates)
        translated = np.array([point[0] - centroid.x, point[1] - centroid.y])

        # Apply rotation
        rotated_point = np.dot(rotation_matrix, translated)

        # Apply numerical precision fix - round values very close to zero to exactly zero
        rotated_point = np.where(np.abs(rotated_point) < presition_threshold, 0, rotated_point)

        # Translate back (add centroid coordinates)
        final_point = (rotated_point[0] + centroid.x, rotated_point[1] + centroid.y)

        # Append to the list of rotated coordinates
        rotated_coords.append(final_point)

    # Create and return a new LinearRing with the rotated coordinates
    return LinearRing(rotated_coords)



"""Formula 9.8N from EN 1992-1-1:2004: Chapter 9 - Detailing of members and particular rules."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MM
from blueprints.validations import raise_if_negative


class Form9Dot8nMaximumTransverseDistanceLegsSeriesShearLinks(Formula):
    """Class representing the formula 9.8N for the calculation of the maximum distance in transverse direction between legs in a series of shear
    links.
    """

    label = "9.8N"
    source_document = EN_1992_1_1_2004

    def __init__(self, d: MM) -> None:
        r"""[$s_{t,max}$] Maximum distance in transverse direction between legs in a series of shear links [mm].

        EN 1992-1-1:2004 art.9.2.2(8) - Formula (9.8N)

        Parameters
        ----------
        d: MM
            [$d$] Effective height of the cross-section [mm].
        """
        super().__init__()
        self.d = d

    @staticmethod
    def _evaluate(d: MM) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(d=d)
        return min(0.75 * d, 600)

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 9.8N."""
        return LatexFormula(
            return_symbol=r"s_{t,max}",
            result=f"{self:.{n}f}",
            equation=r"min(0.75 \cdot d, 600 \text{mm})",
            numeric_equation=rf"min(0.75 \cdot {self.d:.{n}f}, 600 \text{{mm}})",
            comparison_operator_label="=",
        )



"""Testing formula 7.1 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_1 import Form7Dot1MinReinforcingSteel
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot1MinReinforcingSteel:
    """Validation for formula 7.1 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        k_c = 1.0
        k = 0.65
        f_ct_eff = 2.9
        a_ct = 1000.0
        sigma_s = 435.0

        # Object to test
        formula = Form7Dot1MinReinforcingSteel(k_c=k_c, k=k, f_ct_eff=f_ct_eff, a_ct=a_ct, sigma_s=sigma_s)

        # Expected result, manually calculated
        manually_calculated_result = 4.333  # mm^2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("k_c", "k", "f_ct_eff", "a_ct", "sigma_s"),
        [
            (-1.0, 0.65, 2.9, 1000.0, 435.0),  # k_c is negative
            (1.0, -0.65, 2.9, 1000.0, 435.0),  # k is negative
            (1.0, 0.65, -2.9, 1000.0, 435.0),  # f_ct_eff is negative
            (1.0, 0.65, 2.9, -1000.0, 435.0),  # a_ct is negative
            (1.0, 0.65, 2.9, 1000.0, -435.0),  # sigma_s is negative
            (1.0, 0.65, 2.9, 1000.0, 0.0),  # sigma_s is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, k_c: float, k: float, f_ct_eff: float, a_ct: float, sigma_s: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot1MinReinforcingSteel(k_c=k_c, k=k, f_ct_eff=f_ct_eff, a_ct=a_ct, sigma_s=sigma_s)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"A_{s,min} = \frac{k_c \cdot k \cdot f_{ct,eff} \cdot A_{ct}}{\sigma_s} = "
                r"\frac{1.000 \cdot 0.650 \cdot 2.900 \cdot 1000.000}{435.000} = 4.333 \ mm^2",
            ),
            ("short", r"A_{s,min} = 4.333 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k_c = 1.0
        k = 0.65
        f_ct_eff = 2.9
        a_ct = 1000.0
        sigma_s = 435.0

        # Object to test
        latex = Form7Dot1MinReinforcingSteel(k_c=k_c, k=k, f_ct_eff=f_ct_eff, a_ct=a_ct, sigma_s=sigma_s).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 3.19 and 3.20 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA


class Form3Dot19And20EffectivePressureZoneHeight(Formula):
    """Class representing formula 3.19 and 3.20 for the calculation of the λ factor for the effective pressure zone height."""

    label = "3.19 - 3.20"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        r"""[$\lambda$] Factor effective pressure zone height [$-$].

        EN 1992-1-1:2004 art.3.1.7(3) - Formula (3.19) and (3.20)

        Parameters
        ----------
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength concrete [$MPa$].
            Valid range: [$f_{ck} \leq 90$].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> DIMENSIONLESS:
        r"""Evaluates the formula, for more information see the __init__ method."""
        if f_ck <= 50:
            return 0.8
        if f_ck <= 90:
            return 0.8 - (f_ck - 50) / 400
        raise ValueError(f"Invalid f_ck: {f_ck}. Maximum of f_ck is 90 MPa")

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.19 and 3.20."""
        return LatexFormula(
            return_symbol=r"\lambda",
            result=f"{self:.{n}f}",
            equation=r"0.8" if self.f_ck <= 50 else r"0.8 - (f_{ck} - 50) / 400",
            numeric_equation=r"0.8" if self.f_ck <= 50 else rf"0.8 - ({self.f_ck:.{n}f} - 50) / 400",
            comparison_operator_label="=",
        )



"""Testing formula 6.64 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_64 import Form6Dot64BondFactor
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot64BondFactor:
    """Validation for formula 6.64 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_s = 1500.0
        a_p = 1200.0
        xi = 0.8
        d_s = 25.0
        d_p = 15.0

        # Object to test
        formula = Form6Dot64BondFactor(a_s=a_s, a_p=a_p, xi=xi, d_s=d_s, d_p=d_p)

        # Expected result, manually calculated
        manually_calculated_result = 0.93566744136  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_s", "a_p", "xi", "d_s", "d_p"),
        [
            (-1500.0, 1200.0, 0.8, 25.0, 15.0),  # a_s is negative
            (1500.0, -1200.0, 0.8, 25.0, 15.0),  # a_p is negative
            (1500.0, 1200.0, -0.8, 25.0, 15.0),  # xi is negative
            (1500.0, 1200.0, 0.8, -25.0, 15.0),  # d_s is negative
            (1500.0, 1200.0, 0.8, 25.0, -15.0),  # d_p is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_s: float, a_p: float, xi: float, d_s: float, d_p: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot64BondFactor(a_s=a_s, a_p=a_p, xi=xi, d_s=d_s, d_p=d_p)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\eta = \frac{A_s + A_P}{A_s + A_P \cdot \sqrt{\xi \cdot ⌀_s / ⌀_P}} = "
                r"\frac{1500.000 + 1200.000}{1500.000 + 1200.000 \cdot \sqrt{0.800 \cdot 25.000 / 15.000}} = 0.936 \ -",
            ),
            ("short", r"\eta = 0.936 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_s = 1500.0
        a_p = 1200.0
        xi = 0.8
        d_s = 25.0
        d_p = 15.0

        # Object to test
        latex = Form6Dot64BondFactor(a_s=a_s, a_p=a_p, xi=xi, d_s=d_s, d_p=d_p).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.17 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_17 import Form5Dot17EffectiveLengthBucklingLoad
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot17EffectiveLengthBucklingLoad:
    """Validation for formula 5.17 from EN 1992-1-1:2004."""

    @pytest.fixture
    def form_5_17(self) -> Form5Dot17EffectiveLengthBucklingLoad:
        """Setup and teardown for test."""
        return Form5Dot17EffectiveLengthBucklingLoad(ei=1_000_000, n_b=5)

    def test_evaluation(self, form_5_17: Form5Dot17EffectiveLengthBucklingLoad) -> None:
        """Test the evaluation of the result."""
        # Expected result, manually calculated
        manually_calculated_result = 1404.96  # M
        assert form_5_17 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("ei", "n_b"),
        [
            (-1_000_000, 5),
            (1_000_000, -5),
            (-1_000_000, -5),
        ],
    )
    def test_raise_error_when_zero_pars_are_given(self, ei: float, n_b: float) -> None:
        """Test zero values for ei, n_b."""
        with pytest.raises(NegativeValueError):
            Form5Dot17EffectiveLengthBucklingLoad(ei=ei, n_b=n_b)

    def test_raise_error_when_negative_pars_are_given(self) -> None:
        """Test negative values for ei, n_b."""
        ei = 1_000_000
        n_b = 0
        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot17EffectiveLengthBucklingLoad(ei=ei, n_b=n_b)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"l_0 = \pi \cdot \sqrt{\frac{EI}{N_{b}}} = \pi \cdot \sqrt{\frac{1000000}{5}} = 1404.963",
            ),
            ("short", r"l_0 = 1404.963"),
        ],
    )
    def test_latex(self, form_5_17: Form5Dot17EffectiveLengthBucklingLoad, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Object to test
        form_5_17_latex = form_5_17.latex()
        actual = {"complete": form_5_17_latex.complete, "short": form_5_17_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Formula 6.45 from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

import numpy as np

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import MM, MM2
from blueprints.validations import raise_if_negative


class Form6Dot45W1Rectangular(Formula):
    r"""Class representing formula 6.45 for the calculation of [$W_1$]."""

    label = "6.45"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        c_1: MM,
        c_2: MM,
        d: MM,
    ) -> None:
        r"""[$W_1$] Calculation of [$W_1$].

        EN 1992-1-1:2004 art.6.4.3(4) - Formula (6.45)

        Parameters
        ----------
        c_1 : MM
            [$c_1$] Column dimension parallel to the eccentricity of the load [$mm$].
        c_2 : MM
            [$c_2$] Column dimension perpendicular to the eccentricity of the load [$mm$].
        d : MM
            [$d$] Mean effective depth of the slab [$mm$].
        """
        super().__init__()
        self.c_1 = c_1
        self.c_2 = c_2
        self.d = d

    @staticmethod
    def _evaluate(
        c_1: MM,
        c_2: MM,
        d: MM,
    ) -> MM2:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(c_1=c_1, c_2=c_2, d=d)

        return (c_2**2) / 4 + c_1 * c_2 + 4 * c_1 * d + 8 * d**2 + np.pi * d * c_2

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.45."""
        _equation: str = r"\frac{c_2^2}{4} + c_1 \cdot c_2 + 4 \cdot c_1 \cdot d + 8 \cdot d^2 + \pi \cdot d \cdot c_2"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"c_1": f"{self.c_1:.{n}f}",
                r"c_2": f"{self.c_2:.{n}f}",
                r" d": f" {self.d:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"W_1",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="mm^2",
        )



r"""Calculation of nominal concrete cover from EN 1992-1-1: Chapter 4 - Durability and cover to reinforcement."""

from dataclasses import dataclass, field
from typing import Literal

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.nominal_cover_constants import (
    AbrasionClass,
    CastingSurface,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.nominal_cover_constants import (
    NominalConcreteCoverConstantsBase as ConstantsBase,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover._base_classes.structural_class import ConcreteStructuralClassBase
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.formula_4_1 import Form4Dot1NominalConcreteCover
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.formula_4_2 import Form4Dot2MinimumConcreteCover
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.table_4_1 import (
    Carbonation,
    Chemical,
    Chloride,
    ChlorideSeawater,
    FreezeThaw,
    Table4Dot1ExposureClasses,
)
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.table_4_2 import Table4Dot2MinimumCoverWithRegardToBond
from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_4_durability_and_cover.table_4_4n import (
    Table4Dot4nMinimumCoverDurabilityReinforcementSteel,
)
from blueprints.codes.latex_formula import latex_max_curly_brackets
from blueprints.type_alias import MM


@dataclass(frozen=True)
class NominalConcreteCover:
    r"""Class responsible for the calculation of the nominal concrete cover [$c_{nom}$] [$mm$].
    It takes considerations of art.4.4.1.2 and 4.4.1.3 into account.

    Parameters
    ----------
    reinforcement_diameter: MM
        The diameter of the reinforcement [$mm$].
    nominal_max_aggregate_size: MM
        The nominal maximum aggregate size [$mm$].
    constants: ConstantsBase
        The constants for the calculation of the nominal concrete cover.
    structural_class: ConcreteStructuralClassBase | int
        The structural class of the concrete. Either an instance of the ConcreteStructuralClassBase class or an integer.
        Tip: Use the :class:`Table4Dot3ConcreteStructuralClass` class to calculate the structural class.
    carbonation: Carbonation | Literal["XC1", "XC2", "XC3", "XC4", "NA"]
        The classification of corrosion induced by carbonation. Default is "Not applicable".
    chloride: Chloride | Literal["XD1", "XD2", "XD3", "NA"]
        The classification of corrosion induced by chlorides other than by sea water. Default is "Not applicable".
    chloride_seawater: ChlorideSeawater | Literal["XS1", "XS2", "XS3", "NA"]
        The classification of corrosion induced by chlorides from sea water. Default is "Not applicable".
    delta_c_dur_gamma: MM
        [$\Delta c_{dur,\gamma}$] An additional safety requirement based on art. 4.4.1.2 (6) [$mm$].
        The value of [$\Delta c_{dur,\gamma}$] for use in a Country may be found in its National Annex.
        The recommended value is 0 mm. 0 mm is the default value in the formula if not specified otherwise.
    delta_c_dur_st: MM
        [$\Delta c_{dur,st}$] A reduction of minimum concrete cover when using stainless steel based on art. 4.4.1.2 (7) [$mm$].
        The value of [$\Delta c_{dur,st}$] for use in a Country may be found in its National Annex.
        The recommended value, without further specification, is 0 mm. 0 mm is the default value in the formula if not specified otherwise.
    delta_c_dur_add: MM
        [$\Delta c_{dur,add}$] A reduction of minimum concrete cover when using additional protection based on art. 4.4.1.2 (8) [$mm$].
        The value of [$\Delta c_{dur,add}$] for use in a Country may be found in its National Annex.
        The recommended value, without further specification, is 0 mm. 0 mm is the default value in the formula if not specified otherwise.
    casting_surface: CastingSurface
        The casting surface of the concrete according to art. 4.4.1.3 (4).
        The default value is "Permanently exposed".
    uneven_surface: bool
        Is the surface uneven according to art. 4.4.1.2 (11)?
        The default value is False.
    abrasion_class: AbrasionClass
        The abrasion class of the concrete surface according to art. 4.4.1.2 (13).
        The default value is "Not applicable".
    """

    label = "Nominal concrete cover according to art. 4.4.1"
    source_document = "EN 1992-1-1"

    reinforcement_diameter: MM
    nominal_max_aggregate_size: MM
    constants: ConstantsBase
    structural_class: ConcreteStructuralClassBase | int
    carbonation: Carbonation | Literal["XC1", "XC2", "XC3", "XC4", "NA"] = field(default_factory=lambda: Carbonation.NA)
    chloride: Chloride | Literal["XD1", "XD2", "XD3", "NA"] = field(default_factory=lambda: Chloride.NA)
    chloride_seawater: ChlorideSeawater | Literal["XS1", "XS2", "XS3", "NA"] = field(default_factory=lambda: ChlorideSeawater.NA)
    delta_c_dur_gamma: MM = field(default=0)
    delta_c_dur_st: MM = field(default=0)
    delta_c_dur_add: MM = field(default=0)
    casting_surface: CastingSurface = field(default=CastingSurface.PERMANENTLY_EXPOSED)
    uneven_surface: bool = field(default=False)
    abrasion_class: AbrasionClass = field(default=AbrasionClass.NA)

    def __post_init__(self) -> None:
        """Check the input parameters for validity."""
        if not isinstance(self.uneven_surface, bool):
            raise TypeError(f"Invalid type for uneven_surface: {type(self.uneven_surface)}. Expected type is bool.")

        if not isinstance(self.abrasion_class, AbrasionClass):
            raise TypeError(f"Invalid type for abrasion_class: {type(self.abrasion_class)}. Expected type is AbrasionClass.")

        if not isinstance(self.casting_surface, CastingSurface):
            raise TypeError(f"Invalid type for casting_surface: {type(self.casting_surface)}. Expected type is CastingSurface.")

        if isinstance(self.carbonation, str):
            object.__setattr__(self, "carbonation", Carbonation[self.carbonation.upper()])
        if isinstance(self.chloride, str):
            object.__setattr__(self, "chloride", Chloride[self.chloride.upper()])
        if isinstance(self.chloride_seawater, str):
            object.__setattr__(self, "chloride_seawater", ChlorideSeawater[self.chloride_seawater.upper()])

    def exposure_classes(self) -> Table4Dot1ExposureClasses:
        """Exposure classes according to table 4.1 from EN 1992-1-1."""
        return Table4Dot1ExposureClasses(self.carbonation, self.chloride, self.chloride_seawater, FreezeThaw.NA, Chemical.NA)  # type: ignore[arg-type]

    def c_min_b(self) -> Table4Dot2MinimumCoverWithRegardToBond:
        """Minimum concrete cover with regard to bond according to table 4.2 from EN 1992-1-1."""
        return Table4Dot2MinimumCoverWithRegardToBond(self.reinforcement_diameter, self.nominal_max_aggregate_size > 32)

    def c_min_dur(self) -> Table4Dot4nMinimumCoverDurabilityReinforcementSteel:
        """Minimum concrete cover with regard to durability according to table 4.4N from EN 1992-1-1."""
        return Table4Dot4nMinimumCoverDurabilityReinforcementSteel(self.exposure_classes(), self.structural_class)  # type: ignore[arg-type]

    def c_min(self) -> Form4Dot2MinimumConcreteCover:
        """Minimum concrete cover according to formula 4.2 from EN 1992-1-1."""
        return Form4Dot2MinimumConcreteCover(
            c_min_b=self.c_min_b(),
            c_min_dur=self.c_min_dur(),
            delta_c_dur_gamma=self.delta_c_dur_gamma,
            delta_c_dur_st=self.delta_c_dur_st,
            delta_c_dur_add=self.delta_c_dur_add,
        )

    def cover_increase_for_uneven_surface(self) -> MM:
        """Calculate the increase of the concrete cover for uneven surface according to art. 4.4.1.2 (11)."""
        return self.constants.COVER_INCREASE_FOR_UNEVEN_SURFACE * self.uneven_surface

    def cover_increase_for_abrasion_class(self) -> MM:
        """Calculate the increase of the concrete cover for abrasion class according to art. 4.4.1.2 (13)."""
        return self.constants.COVER_INCREASE_FOR_ABRASION_CLASS[self.abrasion_class]

    def c_min_total(self) -> MM:
        """Total minimum concrete cover according to art. 4.4.1.2 (11) and (13) from EN 1992-1-1."""
        c_min = self.c_min()
        # According to art. 4.4.1.2 (11) from EN 1992-1-1
        c_min += self.cover_increase_for_uneven_surface()  # type: ignore[assignment]
        # According to art. 4.4.1.2 (13) from EN 1992-1-1
        c_min += self.cover_increase_for_abrasion_class()  # type: ignore[assignment]
        return c_min

    def c_nom(self) -> Form4Dot1NominalConcreteCover:
        """Nominal concrete cover according to art. 4.4.1 from EN 1992-1-1."""
        return Form4Dot1NominalConcreteCover(c_min=self.c_min_total(), delta_c_dev=self.constants.DEFAULT_DELTA_C_DEV)

    def minimum_cover_with_regard_to_casting_surface(self) -> MM:
        """Calculate the minimum cover with regard to casting surface according to art. 4.4.1.3 (4) from EN 1992-1-1."""
        return self.constants.minimum_cover_with_regard_to_casting_surface(self.c_min_dur(), self.casting_surface)

    def value(self) -> MM:
        """Get the value of the nominal concrete cover."""
        return max(
            self.c_nom(),
            self.minimum_cover_with_regard_to_casting_surface(),
        )

    def latex(self, n: int = 1) -> str:
        """Returns the lateX string representation for Nominal concrete cover check."""
        min_surface = self.minimum_cover_with_regard_to_casting_surface()
        return r"\newline~".join(
            [
                f"Nominal concrete cover according to art. 4.4.1 from {self.constants.CODE_PREFIX}EN 1992-1-1{self.constants.CODE_SUFFIX}:",
                latex_max_curly_brackets(
                    r"Nominal concrete cover according to art. 4.4.1 (c_{nom})",
                    "Minimum cover with regard to casting surface according to art. 4.4.1.3 (4)",
                ),
                f"= {latex_max_curly_brackets(self.c_nom().latex().result, min_surface)} = {self.value():.{n}f} mm",
                "",
                "Where:",
                f"\\hspace{{4ex}}{self.c_nom().latex().return_symbol} = {self.c_nom().latex().equation.replace('min', 'min,total')}"
                f" = {self.c_nom().latex().numeric_equation} = {self.c_nom().latex().result} mm",
                r"\hspace{4ex}\Delta c_{dev} is determined according to art. 4.4.1.3 (1)",
                r"\hspace{4ex}c_{min,total} = c_{min} + \Delta c_{uneven surface}  + \Delta c_{abrasion class}"
                f" = {self.c_min().latex().result} + {self.cover_increase_for_uneven_surface()} + {self.cover_increase_for_abrasion_class()}"
                f" = {self.c_min_total()} mm",
                r"\hspace{4ex}\Delta c_{uneven surface} and \Delta c_{abrasion class} are determined according to art. 4.4.1.2 (11) and (13)",
                f"\\hspace{{4ex}}{self.c_min().latex().return_symbol} = {self.c_min().latex().equation}"
                f" = {self.c_min().latex().numeric_equation} = {self.c_min().latex().result} mm",
                r"\hspace{4ex}\Delta c_{dur,\gamma} , \Delta c_{dur,st} and \Delta c_{dur,add} "
                r"are determined according to art. 4.4.1.2 (6), (7) and (8)",
                f"\\hspace{{4ex}}{self.c_min_b().latex().return_symbol} is determined according to "
                f"table 4.2 based on {self.c_min_b().latex().equation}"
                f" = {self.c_min_b().latex().numeric_equation} = {self.c_min_b().latex().result} mm",
                f"\\hspace{{4ex}}{self.c_min_dur().latex().return_symbol} is determined according to "
                f"table 4.4 based on {self.c_min_dur().latex().equation}"
                f" = {self.c_min_dur().latex().result} mm",
                r"\hspace{4ex}Minimum cover with regard to casting surface according to art. 4.4.1.3 (4) = "
                + self.constants.minimum_cover_with_regard_to_casting_surface_latex(self.casting_surface),
            ]
        ).replace(" ", "~")

    def __str__(self) -> str:
        """Return the string representation of the nominal concrete cover."""
        return f"{self.label} = {self.value()} mm"



"""Testing formula 6.28 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_28 import Form6Dot28RequiredCrossSectionalArea
from blueprints.validations import GreaterThan90Error, LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot28RequiredCrossSectionalArea:
    """Validation for formula 6.28 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        u_k = 500.0
        f_yd = 435.0
        t_ed = 20.0
        a_k = 1000.0
        theta = 30.0

        # Object to test
        formula = Form6Dot28RequiredCrossSectionalArea(u_k=u_k, f_yd=f_yd, t_ed=t_ed, a_k=a_k, theta=theta)

        # Expected result, manually calculated
        manually_calculated_result = 0.01990862997  # mm2

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("u_k", "f_yd", "t_ed", "a_k", "theta"),
        [
            (-500.0, 435.0, 20.0, 1000.0, 30.0),  # u_k is negative
            (500.0, -435.0, 20.0, 1000.0, 30.0),  # f_yd is negative
            (500.0, 435.0, -20.0, 1000.0, 30.0),  # t_ed is negative
            (500.0, 435.0, 20.0, -1000.0, 30.0),  # a_k is negative
            (500.0, 435.0, 20.0, 1000.0, -30.0),  # theta is negative
            (500.0, 0.0, 20.0, 1000.0, 30.0),  # f_yd is zero
            (500.0, 435.0, 20.0, 0.0, 30.0),  # a_k is zero
            (500.0, 435.0, 20.0, 1000.0, 0.0),  # theta is zero
            (500.0, 435.0, 20.0, 1000.0, 91.0),  # theta is greater than 90
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, u_k: float, f_yd: float, t_ed: float, a_k: float, theta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError, GreaterThan90Error)):
            Form6Dot28RequiredCrossSectionalArea(u_k=u_k, f_yd=f_yd, t_ed=t_ed, a_k=a_k, theta=theta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\Sigma A_{sl} = \frac{u_k}{f_{yd}} \cdot \frac{T_{Ed}}{2 \cdot A_k} \cdot \cot(\theta) = "
                r"\frac{500.000}{435.000} \cdot \frac{20.000}{2 \cdot 1000.000} \cdot \cot(30.000) = 0.020 \ mm^2",
            ),
            ("short", r"\Sigma A_{sl} = 0.020 \ mm^2"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        u_k = 500.0
        f_yd = 435.0
        t_ed = 20.0
        a_k = 1000.0
        theta = 30.0

        # Object to test
        latex = Form6Dot28RequiredCrossSectionalArea(u_k=u_k, f_yd=f_yd, t_ed=t_ed, a_k=a_k, theta=theta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 5.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_5_structural_analysis.formula_5_2 import Form5Dot2Eccentricity
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm5Dot2Eccentricity:
    """Validation for formula 5.2 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        theta_i = 0.003  # -
        l_0 = 5  # m
        form_5_2 = Form5Dot2Eccentricity(theta_i=theta_i, l_0=l_0)

        # Expected result, manually calculated
        manually_calculated_result = 0.0075

        assert form_5_2 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_theta_i_is_given(self) -> None:
        """Test a negative value for theta_i."""
        # Example values
        theta_i = -0.003  # -
        l_0 = 5  # m

        with pytest.raises(NegativeValueError):
            Form5Dot2Eccentricity(theta_i=theta_i, l_0=l_0)

    def test_raise_error_when_negative_l_0_is_given(self) -> None:
        """Test a negative value for l_0."""
        # Example values
        theta_i = 0.003  # -
        l_0 = -5  # m

        with pytest.raises(LessOrEqualToZeroError):
            Form5Dot2Eccentricity(theta_i=theta_i, l_0=l_0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"e_i = \theta_i \cdot l_0 / 2 = 0.003 \cdot 5.000 / 2 = 0.0075",
            ),
            ("short", r"e_i = 0.0075"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        theta_i = 0.003  # -
        l_0 = 5  # m

        # Object to test
        form_5_2_latex = Form5Dot2Eccentricity(theta_i=theta_i, l_0=l_0).latex()

        actual = {
            "complete": form_5_2_latex.complete,
            "short": form_5_2_latex.short,
        }

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.16 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_17 import Form6Dot17NominalWebWidth
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot16NominalWebWidth:
    """Validation for formula 6.14 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        b_w = 300.0
        diameters = [16.0, 20.0, 25.0]

        # Object to test
        formula = Form6Dot17NominalWebWidth(b_w=b_w, diameters=diameters)

        # Expected result, manually calculated
        manually_calculated_result = 226.8  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("b_w", "diameters"),
        [
            (-300.0, [16.0, 20.0, 25.0]),  # b_w is negative
            (300.0, [-16.0, 20.0, 25.0]),  # one diameter is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, b_w: float, diameters: list[float]) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot17NominalWebWidth(b_w=b_w, diameters=diameters)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"b_{w,nom} = b_{w} - 1.2 \cdot \sum(⌀) = 300.000 - 1.2 \cdot \left(16.000 + 20.000 + 25.000 \right) = 226.800 \ mm",
            ),
            ("short", r"b_{w,nom} = 226.800 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        b_w = 300.0
        diameters = [16.0, 20.0, 25.0]

        # Object to test
        latex = Form6Dot17NominalWebWidth(b_w=b_w, diameters=diameters).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.54 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_54 import Form6Dot54ControlPerimeter
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot54ControlPerimeter:
    """Validation for formula 6.54 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        beta = 1.1
        v_ed = 200.0
        v_rd_c = 1.5
        d = 500.0

        # Object to test
        formula = Form6Dot54ControlPerimeter(beta=beta, v_ed=v_ed, v_rd_c=v_rd_c, d=d)

        # Expected result, manually calculated
        manually_calculated_result = 0.293333333333333  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("beta", "v_ed", "v_rd_c", "d"),
        [
            (-1.1, 200.0, 1.5, 500.0),  # beta is negative
            (1.1, -200.0, 1.5, 500.0),  # v_ed is negative
            (1.1, 200.0, -1.5, 500.0),  # v_rd_c is negative
            (1.1, 200.0, 1.5, -500.0),  # d is negative
            (1.1, 200.0, 0, 500.0),  # v_rd_c is zero
            (1.1, 200.0, 1.5, 0),  # d is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, beta: float, v_ed: float, v_rd_c: float, d: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot54ControlPerimeter(beta=beta, v_ed=v_ed, v_rd_c=v_rd_c, d=d)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"u_{out,ef} = \frac{\beta \cdot V_{Ed}}{v_{Rd,c} \cdot d} = \frac{1.100 \cdot 200.000}{1.500 \cdot 500.000} = 0.293 \ mm",
            ),
            ("short", r"u_{out,ef} = 0.293 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        beta = 1.1
        v_ed = 200.0
        v_rd_c = 1.5
        d = 500.0

        # Object to test
        latex = Form6Dot54ControlPerimeter(beta=beta, v_ed=v_ed, v_rd_c=v_rd_c, d=d).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 8.1 from EN 1992-1-1:2004: Chapter 8 Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import KN, MM, MPA
from blueprints.unit_conversion import KN_TO_N
from blueprints.validations import raise_if_negative


class Form8Dot1RequiredMinimumMandrelDiameter(Formula):
    """Class representing formula 8.1 for the calculation of the required minimum mandrel diameter if it needs to be checked to avoid
    concrete failure.
    """

    label = "8.1"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_bt: KN,
        a_b: MM,
        diameter: MM,
        f_cd: MPA,
    ) -> None:
        r"""[$\Ø_{m,min}$] minimum mandrel diameter if it needs to be checked to avoid concrete failure [$MM$].

        EN 1992-1-1:2004 art.8.3(3) - Formula (8.1)

        Parameters
        ----------
        f_bt: KN
            [$F_{bt}$] Tensile force from ultimate loads in a bar or group of bars in contact at the start of a bend  [$kN$].
        a_b: MM
            [$a_b$] Half of the centre-to-centre distance between bars (or groups of bars) perpendicular
            to the plane of the bend for a given bar (or group of bars in contact).
            For a bar or group of bars adjacent to the face of the member, [$a_b$] should be taken as the cover plus [$\Ø/2$] [$mm$].
        diameter: MM
            [$\Ø$] Diameter of reinforcing bar [$mm$].
        f_cd: MPA
            [$f_{cd}$] Design value of concrete compressive stress [$MPa$].
            Note: The value of [$f_{cd}$] should not be taken greater than that for concrete class C55/67.
        """
        super().__init__()
        self.f_bt = f_bt
        self.a_b = a_b
        self.diameter = diameter
        self.f_cd = f_cd

    @staticmethod
    def _evaluate(
        f_bt: KN,
        a_b: MM,
        diameter: MM,
        f_cd: MPA,
    ) -> MM:
        """For more detailed documentation see the class docstring."""
        raise_if_negative(
            f_bt=f_bt,
            a_b=a_b,
            diameter=diameter,
            f_cd=f_cd,
        )
        return f_bt * KN_TO_N * ((1 / a_b) + 1 / (2 * diameter)) / f_cd

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 8.1."""
        return LatexFormula(
            return_symbol=r"Ø_{m,min}",
            result=f"{self:.{n}f}",
            equation=r"\frac{F_{bt} \left( \frac{1}{a_b} + \frac{1}{2 \cdot Ø} \right) }{f_{cd}}",
            numeric_equation=(
                rf"\frac{{{self.f_bt:.{n}f} \cdot 1000 \cdot \left( \frac{{1}}{{{self.a_b:.{n}f}}}"
                rf" + \frac{{1}}{{2 \cdot {self.diameter:.{n}f}}} \right)}}{{{self.f_cd:.{n}f}}}"
            ),
            comparison_operator_label="=",
        )



"""Test the LNP enum."""

from blueprints.structural_sections.steel.steel_cross_sections.standard_profiles.lnp import LNP


class TestLNP:
    """Tests for the LNP enum."""

    def test_enum_values(self) -> None:
        """Test that enum values are correctly defined."""
        assert LNP.LNP_40x40x4.value == ("LNP 40x40x4", 40, 40, 4, 6, 3)
        assert LNP.LNP_200x100x16.value == ("LNP 200x100x16", 200, 100, 16, 15, 7.5)

    def test_enum_membership(self) -> None:
        """Test that specific values are members of the enum."""
        assert "LNP 40x40x4" in [e.value[0] for e in LNP]
        assert "LNP 200x100x16" in [e.value[0] for e in LNP]

    def test_enum_uniqueness(self) -> None:
        """Test that all enum values are unique."""
        values = [e.value for e in LNP]
        assert len(values) == len(set(values))

    def test_enum_attributes(self) -> None:
        """Test that enum attributes are correctly assigned."""
        profile = LNP.LNP_60x40x7
        assert profile.alias == "LNP 60x40x7"
        assert profile.height == 60
        assert profile.width == 40
        assert profile.base_thickness == 7
        assert profile.web_thickness == 7
        assert profile.root_radius == 6
        assert profile.back_radius == 0
        assert profile.toe_radius == 3



"""Testing formula 3.16 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_16 import Form3Dot16DesignValueTensileStrength


class TestForm3Dot16DesignValueTensileStrength:
    """Validation for formula 3.16 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example values
        alpha_ct = 1.0  # -
        f_ctk_0_05 = 10.5  # MPa
        gamma_c = 0.8

        form_3_16 = Form3Dot16DesignValueTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

        # Expected result, manually calculated
        manually_calculated_result = 13.125

        assert form_3_16 == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_raise_error_when_negative_alpha_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        alpha_ct = -1.0  # -
        f_ctk_0_05 = 10.5  # MPa
        gamma_c = 0.8

        with pytest.raises(ValueError):
            Form3Dot16DesignValueTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

    def test_raise_error_when_negative_f_ctk_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        alpha_ct = 1.0  # -
        f_ctk_0_05 = -10.5  # MPa
        gamma_c = 0.8

        with pytest.raises(ValueError):
            Form3Dot16DesignValueTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

    def test_raise_error_when_negative_gamma_c_is_given(self) -> None:
        """Test a negative value."""
        # Example values
        alpha_ct = 1.0  # -
        f_ctk_0_05 = 10.5  # MPa
        gamma_c = -0.8

        with pytest.raises(ValueError):
            Form3Dot16DesignValueTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"f_{ctd} = \alpha_{ct} \cdot f_{ctk,0.05} / \gamma_C = 1.000 \cdot 10.500 / 0.800 = 13.125",
            ),
            ("short", r"f_{ctd} = 13.125"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        alpha_ct = 1.0  # -
        f_ctk_0_05 = 10.5  # MPa
        gamma_c = 0.8

        # Object to test
        form_3_16_latex = Form3Dot16DesignValueTensileStrength(alpha_ct=alpha_ct, f_ctk_0_05=f_ctk_0_05, gamma_c=gamma_c).latex()

        actual = {"complete": form_3_16_latex.complete, "short": form_3_16_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.21 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_21 import Form6Dot21ShearStressIOrHSection
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot21ShearStressIOrHSection:
    """Validation for formula 6.21 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        v_ed = 1000.0
        a_w = 200.0
        a_f = 150.0

        # Object to test
        formula = Form6Dot21ShearStressIOrHSection(v_ed=v_ed, a_w=a_w, a_f=a_f)

        # Expected result, manually calculated
        manually_calculated_result = 5.0  # MPa

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("v_ed", "a_w", "a_f"),
        [
            (-1000.0, 200.0, 150.0),  # v_ed is negative
            (1000.0, 200.0, -150.0),  # a_f is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, v_ed: float, a_w: float, a_f: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot21ShearStressIOrHSection(v_ed=v_ed, a_w=a_w, a_f=a_f)

    @pytest.mark.parametrize(
        ("v_ed", "a_w", "a_f"),
        [
            (1000.0, 0.0, 150.0),  # a_w is zero
            (1000.0, -200.0, 150.0),  # a_w is negative
        ],
    )
    def test_raise_error_when_a_w_is_invalid(self, v_ed: float, a_w: float, a_f: float) -> None:
        """Test invalid values for a_w."""
        with pytest.raises(LessOrEqualToZeroError):
            Form6Dot21ShearStressIOrHSection(v_ed=v_ed, a_w=a_w, a_f=a_f)

    @pytest.mark.parametrize(
        ("v_ed", "a_w", "a_f"),
        [
            (1000.0, 200.0, 50.0),  # a_f / a_w < 0.6
        ],
    )
    def test_raise_error_when_a_f_divided_by_a_w_is_invalid(self, v_ed: float, a_w: float, a_f: float) -> None:
        """Test invalid values for a_f / a_w."""
        with pytest.raises(ValueError):
            Form6Dot21ShearStressIOrHSection(v_ed=v_ed, a_w=a_w, a_f=a_f)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\tau_{Ed} = \frac{V_{Ed}}{A_w} \text{ if } A_f / A_w \ge 0.6 = "
                r"\frac{1000.000}{200.000} \text{ if } 150.000 / 200.000 \ge 0.6 = 5.000 \ MPa",
            ),
            ("short", r"\tau_{Ed} = 5.000 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        v_ed = 1000.0
        a_w = 200.0
        a_f = 150.0

        # Object to test
        latex = Form6Dot21ShearStressIOrHSection(v_ed=v_ed, a_w=a_w, a_f=a_f).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



r"""Formula 8.16 from EN 1992-1-1:2004: Chapter 8: Detailing of reinforcement and prestressing tendons."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MM, MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form8Dot16BasicTransmissionLength(Formula):
    r"""Class representing formula 8.16 for the calculation of the basic transmission length [$l_{pt}$]."""

    label = "8.16"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        diameter: MM,
        sigma_pm0: MPA,
        f_bpt: MPA,
    ) -> None:
        r"""[$l_{pt}$] Basic value of the transmission length [$mm$].

        EN 1992-1-1:2004 art.8.10.2.2(2) - Formula (8.16)

        Parameters
        ----------
        alpha_1 : DIMENSIONLESS
            [$α_{1}$] Coefficient taking account of the type of release [$-$].

            = 1.0 for gradual release;

            = 1.25 for sudden release.

            Use your own implementation for this value or use :class:`SubForm8Dot16Alpha1` class.
        alpha_2 : DIMENSIONLESS
            [$α_{2}$] Coefficient taking account of the type of prestressing steel [$-$].

            = 0.25 for tendons with circular cross-section;

            = 0.19 for 3 and 7-wire strands.

            Use your own implementation for this value or use :class:`SubForm8Dot16Alpha2` class.
        diameter : MM
            [$Ø$] Nominal diameter of the tendon [$mm$].
        sigma_pm0 : MPA
            [$σ_{pm0}$] Tendon stress at time of release [$MPa$].
        f_bpt : MPA
            [$f_{bpt}$] Constant bond stress at which prestress is assumed to be transferred to the concrete [$MPa$]

            Use your own implementation for this value or use :class:`Form8Dot15PrestressTransferStress` class.
        """
        super().__init__()
        self.alpha_1 = alpha_1
        self.alpha_2 = alpha_2
        self.diameter = diameter
        self.sigma_pm0 = sigma_pm0
        self.f_bpt = f_bpt

    @staticmethod
    def _evaluate(
        alpha_1: DIMENSIONLESS,
        alpha_2: DIMENSIONLESS,
        diameter: MM,
        sigma_pm0: MPA,
        f_bpt: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(
            alpha_1=alpha_1,
            alpha_2=alpha_2,
            diameter=diameter,
            sigma_pm0=sigma_pm0,
        )
        raise_if_less_or_equal_to_zero(f_bpt=f_bpt)
        return alpha_1 * alpha_2 * diameter * sigma_pm0 / f_bpt

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for formula 8.16."""
        return LatexFormula(
            return_symbol=r"l_{pt}",
            result=f"{self:.{n}f}",
            equation=r"\alpha_1 \cdot \alpha_2 \cdot Ø \cdot \frac{\sigma_{pm0}}{f_{bpt}}",
            numeric_equation=(
                rf"{self.alpha_1:.{n}f} \cdot {self.alpha_2:.{n}f} "
                rf"\cdot {self.diameter:.{n}f} \cdot \frac{{{self.sigma_pm0:.{n}f}}}{{{self.f_bpt:.{n}f}}}"
            ),
            comparison_operator_label="=",
        )


class SubForm8Dot16Alpha1(Formula):
    r"""Class representing sub-formula 8.16 for the calculation of the coefficient [$α_{1}$]."""

    label = "8.16"
    source_document = EN_1992_1_1_2004

    def __init__(self, release_type: str) -> None:
        r"""[$α_{1}$] Coefficient taking account of the type of release [$-$].

        EN 1992-1-1:2004 art.8.10.2.2(2) - Formula (8.16)

        Parameters
        ----------
        release_type : str
            Type of release, either "gradual" or "sudden".
        """
        super().__init__()
        self.release_type = release_type

    @staticmethod
    def _evaluate(release_type: str) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        match release_type.lower():
            case "gradual":
                return 1.0
            case "sudden":
                return 1.25
            case _:
                raise ValueError(f"Invalid release type: {release_type}. Valid values are 'gradual' or 'sudden'.")

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for the first subformula of formula 8.16."""
        return LatexFormula(
            return_symbol=r"\alpha_1",
            result=f"{self:.{n}f}",
            equation=r"release\;type",
            numeric_equation=f"{self.release_type}",
            comparison_operator_label=r"\rightarrow",
        )


class SubForm8Dot16Alpha2(Formula):
    r"""Class representing sub-formula 8.16 for the calculation of the coefficient [$α_{2}$]."""

    label = "8.16"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        type_of_wire: str,
    ) -> None:
        r"""[$α_{2}$] Coefficient that takes into account the type of wires in the cross-section [$-$].

        EN 1992-1-1:2004 art.8.10.2.2(2) - Formula (8.16)

        Parameters
        ----------
        type_of_wire : str
            Type of wire.

            = 'circular' for circular cross-sections;

            = '3_7_wire_strands' for 3 and 7-wire strands;
        """
        super().__init__()
        self.type_of_wire = type_of_wire

    @staticmethod
    def _evaluate(type_of_wire: str) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        match type_of_wire.lower():
            case "circular":
                return 0.25
            case "3_7_wire_strands":
                return 0.19
            case _:
                raise ValueError(f"Invalid type of wire: {type_of_wire}. Valid values are 'circular' or '3_7_wire_strands'.")

    def latex(self, n: int = 2) -> LatexFormula:
        """Returns LatexFormula object for the second subformula of formula 8.16."""
        return LatexFormula(
            return_symbol=r"\alpha_2",
            result=f"{self:.{n}f}",
            equation=r"type\;of\;wire",
            numeric_equation=f"{self.type_of_wire}".replace(" ", r"\;"),
            comparison_operator_label=r"\rightarrow",
        )



"""Formula 3.3 from EN 1992-1-1:2004: Chapter 3 - Materials."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA


class Form3Dot3AxialTensileStrengthFromTensileSplittingStrength(Formula):
    """Class representing formula 3.3 for the approximated axial tensile strength, [$f_{ct}$], determined by tensile splitting strength."""

    label = "3.3"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ct_sp: MPA,
    ) -> None:
        r"""[$f_{ct}$] The approximated axial tensile strength when tensile strength is determined as coefficient
        which is dependent of the age of concrete [$MPa$].

        EN 1992-1-1:2004 art.3.1.2(8) - Formula (3.3)

        Parameters
        ----------
        f_ct_sp : MPA
            [$f_{ct,sp}$] Tensile strength determined by tensile splitting strength [$MPa$].

        Returns
        -------
        None
        """
        super().__init__()
        self.f_ct_sp = f_ct_sp

    @staticmethod
    def _evaluate(
        f_ct_sp: MPA,
    ) -> MPA:
        """Evaluates the formula, for more information see the __init__ method."""
        if f_ct_sp < 0:
            raise ValueError(f"Negative f_ct_sp: {f_ct_sp}. f_ct_sp cannot be negative")
        return 0.9 * f_ct_sp

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 3.3."""
        return LatexFormula(
            return_symbol=r"f_{ct}",
            result=f"{self:.{n}f}",
            equation=r"0.9 \cdot f_{ct,sp}",
            numeric_equation=rf"0.9 \cdot {self.f_ct_sp:.{n}f}",
            comparison_operator_label="=",
        )



"""Formula 6.70 from EN 1992-1-1:2004: Chapter 6 - Ultimate limit state."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import MPA
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_lists_differ_in_length, raise_if_negative


class Form6Dot70FatigueDamageFactor(Formula):
    """Class representing formula 6.70 for the calculation of the fatigue damage factor [$D_{Ed}$]."""

    label = "6.70"
    source_document = EN_1992_1_1_2004

    def __init__(self, n_delta_sigma_i: list[MPA], capital_n_delta_sigma_i: list[MPA]) -> None:
        r"""[D_{Ed}] The calculation of the fatigue damage factor [-].

        EN 1992-1-1:2004 art.6.8.4(2) - Formula (6.70)

        Parameters
        ----------
        n_delta_sigma_i : list[MPA]
            [$n(\Delta \sigma_i)$] The applied number of cycles for a stress range [MPa].
        capital_n_delta_sigma_i : list[MPA]
            [$N(\Delta \sigma_i)$] The resisting number of cycles for a stress range [MPa]

        Returns
        -------
        None
        """
        super().__init__()
        self.n_delta_sigma_i = n_delta_sigma_i
        self.capital_n_delta_sigma_i = capital_n_delta_sigma_i

    @staticmethod
    def _evaluate(
        n_delta_sigma_i: list[MPA],
        capital_n_delta_sigma_i: list[MPA],
    ) -> bool:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_lists_differ_in_length(n_delta_sigma_i=n_delta_sigma_i, capital_n_delta_sigma_i=capital_n_delta_sigma_i)
        raise_if_negative(min_n_delta_sigma_i=min(n_delta_sigma_i))
        raise_if_less_or_equal_to_zero(min_capital_n_delta_sigma_i=min(capital_n_delta_sigma_i))
        return sum(n / capital_n for n, capital_n in zip(n_delta_sigma_i, capital_n_delta_sigma_i)) < 1

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.70."""
        _equation: str = r"\sum_{i} \frac{n(\Delta \sigma_i)}{N(\Delta \sigma_i)} < 1"
        _numeric_equation: str = ""
        for n_idx, capital_n in zip(self.n_delta_sigma_i, self.capital_n_delta_sigma_i):
            _numeric_equation += f"\\frac{{{n_idx:.{n}f}}}{{{capital_n:.{n}f}}} + "
        _numeric_equation = _numeric_equation[:-3] + " < 1"
        return LatexFormula(
            return_symbol=r"CHECK",
            result="OK" if self.__bool__() else "\\text{Not OK}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="\\to",
            unit="",
        )



"""Formula 7.18 from EN 1992-1-1:2004: Chapter 7 - Serviceability Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS
from blueprints.validations import raise_if_negative


class Form7Dot18DeformationParameter(Formula):
    r"""Class representing formula 7.18 for the calculation of [$\alpha$]."""

    label = "7.18"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        zeta: DIMENSIONLESS,
        alpha_l: DIMENSIONLESS,
        alpha_ll: DIMENSIONLESS,
    ) -> None:
        r"""[$\alpha$] Calculation of the deformation parameter [$\alpha$].

        EN 1992-1-1:2004 art.7.4.3(3) - Formula (7.18)

        Parameters
        ----------
        zeta : DIMENSIONLESS
            [$\zeta$] Distribution coefficient allowing for tension stiffening at a section, calculated with Expression (7.19).
        alpha_l : DIMENSIONLESS
            [$\alpha_{I}$] Value of the parameter calculated for the uncracked condition
        alpha_ll : DIMENSIONLESS
            [$\alpha_{II}$] Value of the parameter calculated for the fully cracked condition.
        .
        """
        super().__init__()
        self.zeta = zeta
        self.alpha_l = alpha_l
        self.alpha_ll = alpha_ll

    @staticmethod
    def _evaluate(
        zeta: DIMENSIONLESS,
        alpha_l: DIMENSIONLESS,
        alpha_ll: DIMENSIONLESS,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(alpha_ll=alpha_ll, alpha_l=alpha_l, zeta=zeta)

        return zeta * alpha_ll + (1 - zeta) * alpha_l

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 7.18."""
        _equation: str = r"\zeta \cdot \alpha_{II} + (1 - \zeta) \cdot \alpha_{I}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"\zeta": f"{self.zeta:.{n}f}",
                r"\alpha_{II}": f"{self.alpha_ll:.{n}f}",
                r"\alpha_{I}": f"{self.alpha_l:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"\alpha",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="-",
        )



"""Testing sub-formula for 3.2 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_3_materials.formula_3_2 import SubForm3Dot2CoefficientTypeOfCementS


class TestSubForm3Dot2CoefficientTypeOfCementS:
    """Validation for formula 3.2 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Test the evaluation of the result."""
        # Example value 1
        cement_class = "R"  # str
        sub_form_3_2 = SubForm3Dot2CoefficientTypeOfCementS(cement_class=cement_class)
        # Expected result, manually calculated
        manually_result = 0.20
        assert sub_form_3_2 == pytest.approx(expected=manually_result, rel=1e-4)

        # Example value 2
        cement_class = "N"  # str
        sub_form_3_2 = SubForm3Dot2CoefficientTypeOfCementS(cement_class=cement_class)
        # Expected result, manually calculated
        manually_result = 0.25
        assert sub_form_3_2 == pytest.approx(expected=manually_result, rel=1e-4)

        # Example value 3
        cement_class = "S"  # str
        sub_form_3_2 = SubForm3Dot2CoefficientTypeOfCementS(cement_class=cement_class)
        # Expected result, manually calculated
        manually_result = 0.38
        assert sub_form_3_2 == pytest.approx(expected=manually_result, rel=1e-4)

    def test_raise_error_when_invalid_cement_class_is_given(self) -> None:
        """Test an invalid cement class."""
        # Example values
        cement_class = "V"  # str

        with pytest.raises(ValueError):
            SubForm3Dot2CoefficientTypeOfCementS(cement_class=cement_class)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"s \rightarrow \text{cement class} \rightarrow R \rightarrow 0.200",
            ),
            ("short", r"s \rightarrow 0.200"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        cement_class = "R"

        # Object to test
        form_7_3_latex = SubForm3Dot2CoefficientTypeOfCementS(cement_class=cement_class).latex()

        actual = {"complete": form_7_3_latex.complete, "short": form_7_3_latex.short}

        assert actual[representation] == expected, f"{representation} representation failed."



"""Testing formula 6.9 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_9 import Form6Dot9MaximumShearResistance
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot9MaximumShearResistance:
    """Validation for formula 6.9 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        b_w = 300.0
        z = 500.0
        f_cd = 25.0
        nu_1 = 0.6
        alpha_cw = 1.0
        theta = 30.0

        # Object to test
        formula = Form6Dot9MaximumShearResistance(b_w=b_w, z=z, f_cd=f_cd, nu_1=nu_1, alpha_cw=alpha_cw, theta=theta)

        # Expected result, manually calculated
        manually_calculated_result = 974278.579257  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("b_w", "z", "f_cd", "nu_1", "alpha_cw", "theta"),
        [
            (-300.0, 500.0, 25.0, 0.6, 1.0, 30.0),  # b_w is negative
            (300.0, -500.0, 25.0, 0.6, 1.0, 30.0),  # z is negative
            (300.0, 500.0, -25.0, 0.6, 1.0, 30.0),  # f_cd is negative
            (300.0, 500.0, 25.0, -0.6, 1.0, 30.0),  # nu_1 is negative
            (300.0, 500.0, 25.0, 0.6, -1.0, 30.0),  # alpha_cw is negative
            (300.0, 500.0, 25.0, 0.6, 1.0, -30.0),  # theta is negative
            (300.0, 500.0, 25.0, 0.6, 1.0, 0.0),  # theta is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, b_w: float, z: float, f_cd: float, nu_1: float, alpha_cw: float, theta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot9MaximumShearResistance(b_w=b_w, z=z, f_cd=f_cd, nu_1=nu_1, alpha_cw=alpha_cw, theta=theta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"V_{Rd,max} = \alpha_{cw} \cdot b_{w} \cdot z \cdot \nu_{1} \cdot \frac{f_{cd}}{\cot(\theta) + \tan(\theta)} = "
                r"1.000 \cdot 300.000 \cdot 500.000 \cdot 0.600 \cdot \frac{25.000}{\cot(30.000) + \tan(30.000)} = 974278.579 \ N",
            ),
            ("short", r"V_{Rd,max} = 974278.579 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        b_w = 300.0
        z = 500.0
        f_cd = 25.0
        nu_1 = 0.6
        alpha_cw = 1.0
        theta = 30.0

        # Object to test
        latex = Form6Dot9MaximumShearResistance(b_w=b_w, z=z, f_cd=f_cd, nu_1=nu_1, alpha_cw=alpha_cw, theta=theta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.24 from EN 1993-1-1:2005: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import NMM
from blueprints.validations import raise_if_negative


class Form6Dot24TotalTorsionalMoment(Formula):
    r"""Class representing formula 6.24 for the calculation of [$T_{Ed}$]."""

    label = "6.24"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        t_t_ed: NMM,
        t_w_ed: NMM,
    ) -> None:
        r"""[$T_{Ed}$] Calculation of the total torsional moment [$Nmm$].

        EN 1993-1-1:2005 art.6.2.7(2) - Formula (6.24)

        Parameters
        ----------
        t_t_ed : NMM
            [$T_{t,Ed}$] Internal St. Venant torsion [$Nmm$].
        t_w_ed : NMM
            [$T_{w,Ed}$] Internal warping torsion [$Nmm$].
        """
        super().__init__()
        self.t_t_ed = t_t_ed
        self.t_w_ed = t_w_ed

    @staticmethod
    def _evaluate(
        t_t_ed: NMM,
        t_w_ed: NMM,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(t_t_ed=t_t_ed, t_w_ed=t_w_ed)

        return t_t_ed + t_w_ed

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.24."""
        _equation: str = r"T_{t,Ed} + T_{w,Ed}"
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"T_{t,Ed}": f"{self.t_t_ed:.{n}f}",
                r"T_{w,Ed}": f"{self.t_w_ed:.{n}f}",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"T_{Ed}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Testing formula 7.10 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_7_serviceability_limit_state.formula_7_10 import Form7Dot10RhoPEff
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm7Dot10RhoPEff:
    """Validation for formula 7.10 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a_s = 1500.0
        xi_1 = 0.7225
        a_p_prime = 1200.0
        a_c_eff = 25000.0

        # Object to test
        formula = Form7Dot10RhoPEff(a_s=a_s, xi_1=xi_1, a_p_prime=a_p_prime, a_c_eff=a_c_eff)

        # Expected result, manually calculated
        manually_calculated_result = 0.09468  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a_s", "xi_1", "a_p_prime", "a_c_eff"),
        [
            (-1500.0, 0.7225, 1200.0, 25000.0),  # a_s is negative
            (1500.0, -0.7225, 1200.0, 25000.0),  # xi_1 is negative
            (1500.0, 0.7225, -1200.0, 25000.0),  # a_p_prime is negative
            (1500.0, 0.7225, 1200.0, -25000.0),  # a_c_eff is negative
            (1500.0, 0.7225, 1200.0, 0.0),  # a_c_eff is zero
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a_s: float, xi_1: float, a_p_prime: float, a_c_eff: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form7Dot10RhoPEff(a_s=a_s, xi_1=xi_1, a_p_prime=a_p_prime, a_c_eff=a_c_eff)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\rho_{p,eff} = \frac{A_s + \xi_1 \cdot A'_p}{A_{c,eff}} = "
                r"\frac{1500.000 + 0.723 \cdot 1200.000}{25000.000} = 0.095 \ -",
            ),
            ("short", r"\rho_{p,eff} = 0.095 \ -"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a_s = 1500.0
        xi_1 = 0.7225
        a_p_prime = 1200.0
        a_c_eff = 25000.0

        # Object to test
        latex = Form7Dot10RhoPEff(a_s=a_s, xi_1=xi_1, a_p_prime=a_p_prime, a_c_eff=a_c_eff).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.30 from NEN-EN 1993-1-1+A1:2016: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1993_1_1_2005 import EN_1993_1_1_2005
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula, latex_replace_symbols
from blueprints.type_alias import DIMENSIONLESS, MM, MM3, MPA, NMM
from blueprints.validations import raise_if_less_or_equal_to_zero, raise_if_negative


class Form6Dot30ReducedPlasticResistanceMoment(Formula):
    r"""Class representing formula 6.30 for the calculation of [$M_{y,V,Rd}$]."""

    label = "6.30"
    source_document = EN_1993_1_1_2005

    def __init__(
        self,
        w_pl_y: MM3,
        rho: DIMENSIONLESS,
        h_w: MM,
        t_w: MM,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        m_y_c_rd: NMM,
    ) -> None:
        r"""[$M_{y,V,Rd}$] Reduced design plastic resistance moment [$Nmm$].

        NEN-EN 1993-1-1+A1:2016 art.6.2.8(5) - Formula (6.30)

        Parameters
        ----------
        w_pl_y : MM3
            [$W_{pl,y}$] Plastic section modulus about the y-axis [$mm^3$].
        rho : DIMENSIONLESS
            [$\rho$] Shear force ratio (see 6.2.8 (3) or equation 6.29 (rho)) [-].
        h_w : MM
            [$h_w$] Web height [$mm$].
        t_w : MM
            [$t_w$] Web thickness [$mm$].
        f_y : MPA
            [$f_y$] Yield strength of the material [$MPa$].
        gamma_m0 : DIMENSIONLESS
            [$\gamma_{M0}$] Partial safety factor for resistance of cross-sections [-].
        m_y_c_rd : NMM
            [$M_{y,c,Rd}$] Design resistance moment, obtained from 6.2.5(2) [$Nmm$].
        """
        super().__init__()
        self.w_pl_y = w_pl_y
        self.rho = rho
        self.h_w = h_w
        self.t_w = t_w
        self.f_y = f_y
        self.gamma_m0 = gamma_m0
        self.m_y_c_rd = m_y_c_rd

    @staticmethod
    def _evaluate(
        w_pl_y: MM3,
        rho: DIMENSIONLESS,
        h_w: MM,
        t_w: MM,
        f_y: MPA,
        gamma_m0: DIMENSIONLESS,
        m_y_c_rd: NMM,
    ) -> NMM:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(w_pl_y=w_pl_y, rho=rho, h_w=h_w, f_y=f_y, m_y_c_rd=m_y_c_rd)
        raise_if_less_or_equal_to_zero(gamma_m0=gamma_m0, t_w=t_w)

        a_w = h_w * t_w
        m_y_v_rd = (w_pl_y - (rho * a_w**2) / (4 * t_w)) * f_y / gamma_m0
        return min(m_y_v_rd, m_y_c_rd)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.30."""
        _equation: str = (
            r"\min\left(\frac{\left[W_{pl,y} - \frac{\rho \cdot (h_w \cdot t_w)^2}{4 \cdot t_w}\right] \cdot f_y}{\gamma_{M0}}, M_{y,c,Rd}\right)"
        )
        _numeric_equation: str = latex_replace_symbols(
            _equation,
            {
                r"W_{pl,y}": f"{self.w_pl_y:.{n}f}",
                r"\rho": f"{self.rho:.{n}f}",
                r"h_w": f"{self.h_w:.{n}f}",
                r"t_w": f"{self.t_w:.{n}f}",
                r"f_y": f"{self.f_y:.{n}f}",
                r"\gamma_{M0}": f"{self.gamma_m0:.{n}f}",
                r"M_{y,c,Rd}": f"{self.m_y_c_rd:.{n}f}",
            },
            False,
        )
        _numeric_equation_with_units: str = latex_replace_symbols(
            _equation,
            {
                r"W_{pl,y}": rf"{self.w_pl_y:.{n}f} \ mm^3",
                r"\rho": rf"{self.rho:.{n}f}",
                r"h_w": rf"{self.h_w:.{n}f} \ mm",
                r"t_w": rf"{self.t_w:.{n}f} \ mm",
                r"f_y": rf"{self.f_y:.{n}f} \ MPa",
                r"\gamma_{M0}": rf"{self.gamma_m0:.{n}f}",
                r"M_{y,c,Rd}": rf"{self.m_y_c_rd:.{n}f} \ Nmm",
            },
            False,
        )
        return LatexFormula(
            return_symbol=r"M_{y,V,Rd}",
            result=f"{self:.{n}f}",
            equation=_equation,
            numeric_equation=_numeric_equation,
            numeric_equation_with_units=_numeric_equation_with_units,
            comparison_operator_label="=",
            unit="Nmm",
        )



"""Testing formula 6.36 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_36 import (
    Form6Dot36ExternalContourRadiusCircularColumnHeads,
)
from blueprints.validations import NegativeValueError


class TestForm6Dot36ExternalContourRadiusCircularColumnHeads:
    """Validation for formula 6.36 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        d = 500.0
        l_h = 1000.0
        c = 300.0

        # Object to test
        formula = Form6Dot36ExternalContourRadiusCircularColumnHeads(d=d, l_h=l_h, c=c)

        # Expected result, manually calculated
        manually_calculated_result = 2150.0  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("d", "l_h", "c"),
        [
            (-500.0, 1000.0, 300.0),  # d is negative
            (500.0, -1000.0, 300.0),  # l_h is negative
            (500.0, 1000.0, -300.0),  # c is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, d: float, l_h: float, c: float) -> None:
        """Test invalid values."""
        with pytest.raises(NegativeValueError):
            Form6Dot36ExternalContourRadiusCircularColumnHeads(d=d, l_h=l_h, c=c)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"r_{cont,ext} = l_{H} + 2 \cdot d + 0.5 \cdot c = 1000.000 + 2 \cdot 500.000 + 0.5 \cdot 300.000 = 2150.000 \ mm",
            ),
            ("short", r"r_{cont,ext} = 2150.000 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        d = 500.0
        l_h = 1000.0
        c = 300.0

        # Object to test
        latex = Form6Dot36ExternalContourRadiusCircularColumnHeads(d=d, l_h=l_h, c=c).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.3N of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_3n import Form6Dot3NShearCapacityWithoutRebar
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot3NShearCapacityWithoutRebar:
    """Validation for formula 6.3N from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        k = 1.0
        f_ck = 30.0

        # Object to test
        formula = Form6Dot3NShearCapacityWithoutRebar(k=k, f_ck=f_ck)

        # Expected result, manually calculated
        manually_calculated_result = 0.19170289512

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("k", "f_ck"),
        [
            (-1.0, 30.0),  # k is negative
            (1.0, -30.0),  # f_ck is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, k: float, f_ck: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot3NShearCapacityWithoutRebar(k=k, f_ck=f_ck)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"v_{min} = 0.035 \cdot k^{3/2} \cdot f_{ck}^{1/2} = 0.035 \cdot 1.000^{3/2} \cdot 30.000^{1/2} = 0.192 \ MPa",
            ),
            ("short", r"v_{min} = 0.192 \ MPa"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        k = 1.0
        f_ck = 30.0

        # Object to test
        latex = Form6Dot3NShearCapacityWithoutRebar(k=k, f_ck=f_ck).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Formula 6.6n from EN 1992-1-1:2004: Chapter 6 - Ultimate Limit State."""

from blueprints.codes.eurocode.en_1992_1_1_2004 import EN_1992_1_1_2004
from blueprints.codes.formula import Formula
from blueprints.codes.latex_formula import LatexFormula
from blueprints.type_alias import DIMENSIONLESS, MPA
from blueprints.validations import raise_if_negative


class Form6Dot6nStrengthReductionFactor(Formula):
    r"""Class representing formula 6.6n for the calculation of the strength reduction factor, [$\nu$]."""

    label = "6.6n"
    source_document = EN_1992_1_1_2004

    def __init__(
        self,
        f_ck: MPA,
    ) -> None:
        r"""[$\nu$] Strength reduction factor for concrete cracked in shear [$-$].

        EN 1992-1-1:2004 art.6.2.2 (6) - Formula (6.6n)

        Parameters
        ----------
        f_ck : MPA
            [$f_{ck}$] Characteristic compressive strength of concrete [$MPa$].
        """
        super().__init__()
        self.f_ck = f_ck

    @staticmethod
    def _evaluate(
        f_ck: MPA,
    ) -> DIMENSIONLESS:
        """Evaluates the formula, for more information see the __init__ method."""
        raise_if_negative(f_ck=f_ck)

        return 0.6 * (1 - f_ck / 250)

    def latex(self, n: int = 3) -> LatexFormula:
        """Returns LatexFormula object for formula 6.6n."""
        return LatexFormula(
            return_symbol=r"\nu",
            result=f"{self:.{n}f}",
            equation=r"0.6 \cdot \left(1 - \frac{f_{ck}}{250}\right)",
            numeric_equation=rf"0.6 \cdot \left(1 - \frac{{{self.f_ck:.{n}f}}}{{250}}\right)",
            comparison_operator_label="=",
            unit="-",
        )



"""Testing formula 6.65 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_65 import Form6Dot65ConcreteCompressionStrut
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot65ConcreteCompressionStrut:
    """Validation for formula 6.65 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        theta = 30.0

        # Object to test
        formula = Form6Dot65ConcreteCompressionStrut(theta=theta)

        # Expected result, manually calculated
        manually_calculated_result = 37.228865915079  # dimensionless

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        "theta",
        [
            -30.0,  # theta is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, theta: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot65ConcreteCompressionStrut(theta=theta)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"\theta_{fat} = \tan^{-1}\left(\min\left(\sqrt{\tan(\theta)}, 1\right)\right)"
                r" = \tan^{-1}\left(\min\left(\sqrt{\tan(30.000)}, 1\right)\right) = 37.229 \ degrees",
            ),
            ("short", r"\theta_{fat} = 37.229 \ degrees"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        theta = 30.0

        # Object to test
        latex = Form6Dot65ConcreteCompressionStrut(theta=theta).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formula 6.10 of EN 1993-1-1:2005."""

import pytest

from blueprints.codes.eurocode.en_1993_1_1_2005.chapter_6_ultimate_limit_state.formula_6_10 import Form6Dot10NcRdClass1And2And3
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot10NcRdClass1And2And3:
    """Validation for formula 6.10 from EN 1993-1-1:2005."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        a = 2000.0
        f_y = 355.0
        gamma_m0 = 1.0

        # Object to test
        formula = Form6Dot10NcRdClass1And2And3(a=a, f_y=f_y, gamma_m0=gamma_m0)

        # Expected result, manually calculated
        manually_calculated_result = 710000.0  # N

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("a", "f_y", "gamma_m0"),
        [
            (-2000.0, 355.0, 1.0),  # a is negative
            (2000.0, -355.0, 1.0),  # f_y is negative
            (2000.0, 355.0, 0.0),  # gamma_m0 is zero
            (2000.0, 355.0, -1.0),  # gamma_m0 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, a: float, f_y: float, gamma_m0: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot10NcRdClass1And2And3(a=a, f_y=f_y, gamma_m0=gamma_m0)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"N_{c,Rd} = \frac{A \cdot f_y}{\gamma_{M0}} = \frac{2000.000 \cdot 355.000}{1.000} = 710000.000 \ N",
            ),
            ("short", r"N_{c,Rd} = 710000.000 \ N"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        a = 2000.0
        f_y = 355.0
        gamma_m0 = 1.0

        # Object to test
        latex = Form6Dot10NcRdClass1And2And3(a=a, f_y=f_y, gamma_m0=gamma_m0).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Testing formulas 6.34 and 6.35 of EN 1992-1-1:2004."""

import pytest

from blueprints.codes.eurocode.en_1992_1_1_2004.chapter_6_ultimate_limit_state.formula_6_34_35 import Form6Dot34And35ContourRadiusRectangular
from blueprints.validations import LessOrEqualToZeroError, NegativeValueError


class TestForm6Dot34And35ContourRadiusRectangular:
    """Validation for formulas 6.34 and 6.35 from EN 1992-1-1:2004."""

    def test_evaluation(self) -> None:
        """Tests the evaluation of the result."""
        # Example values
        d = 500.0
        c_1 = 300.0
        c_2 = 400.0
        l_h1 = 600.0
        l_h2 = 800.0

        # Object to test
        formula = Form6Dot34And35ContourRadiusRectangular(d=d, c_1=c_1, c_2=c_2, l_h1=l_h1, l_h2=l_h2)

        # Expected result, manually calculated
        manually_calculated_result = 1969.94845224  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    def test_evaluation_reversed_values(self) -> None:
        """Tests the evaluation of the result with reversed values."""
        # Example values
        d = 500.0
        c_1 = 400.0
        c_2 = 300.0
        l_h1 = 800.0
        l_h2 = 600.0

        # Object to test
        formula = Form6Dot34And35ContourRadiusRectangular(d=d, c_1=c_1, c_2=c_2, l_h1=l_h1, l_h2=l_h2)

        # Expected result, manually calculated
        manually_calculated_result = 1969.94845224  # mm

        assert formula == pytest.approx(expected=manually_calculated_result, rel=1e-4)

    @pytest.mark.parametrize(
        ("d", "c_1", "c_2", "l_h1", "l_h2"),
        [
            (-500.0, 300.0, 400.0, 600.0, 800.0),  # d is negative
            (500.0, -300.0, 400.0, 600.0, 800.0),  # c_1 is negative
            (500.0, 300.0, -400.0, 600.0, 800.0),  # c_2 is negative
            (500.0, 300.0, 400.0, -600.0, 800.0),  # l_h1 is negative
            (500.0, 300.0, 400.0, 600.0, -800.0),  # l_h2 is negative
        ],
    )
    def test_raise_error_when_invalid_values_are_given(self, d: float, c_1: float, c_2: float, l_h1: float, l_h2: float) -> None:
        """Test invalid values."""
        with pytest.raises((NegativeValueError, LessOrEqualToZeroError)):
            Form6Dot34And35ContourRadiusRectangular(d=d, c_1=c_1, c_2=c_2, l_h1=l_h1, l_h2=l_h2)

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"r_{cont} = min\left(2 \cdot d + 0.56 \cdot \sqrt{(c_1 + 2 \cdot l_{H1}) \cdot (c_2 + 2 \cdot l_{H2})}, "
                r"2 \cdot d + 0.69 \cdot (c_1 + 2 \cdot l_{H1})\right)"
                r" = min\left(2 \cdot 500.000 + 0.56 \cdot \sqrt{(300.000 + 2 \cdot 600.000) \cdot (400.000 + 2 \cdot 800.000)}, "
                r"2 \cdot 500.000 + 0.69 \cdot (300.000 + 2 \cdot 600.000)\right) = 1969.948 \ mm",
            ),
            ("short", r"r_{cont} = 1969.948 \ mm"),
        ],
    )
    def test_latex(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula."""
        # Example values
        d = 500.0
        c_1 = 300.0
        c_2 = 400.0
        l_h1 = 600.0
        l_h2 = 800.0

        # Object to test
        latex = Form6Dot34And35ContourRadiusRectangular(d=d, c_1=c_1, c_2=c_2, l_h1=l_h1, l_h2=l_h2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."

    @pytest.mark.parametrize(
        ("representation", "expected"),
        [
            (
                "complete",
                r"r_{cont} = min\left(2 \cdot d + 0.56 \cdot \sqrt{(c_1 + 2 \cdot l_{H1}) \cdot (c_2 + 2 \cdot l_{H2})}, "
                r"2 \cdot d + 0.69 \cdot (c_1 + 2 \cdot l_{H1})\right)"
                r" = min\left(2 \cdot 500.000 + 0.56 \cdot \sqrt{(300.000 + 2 \cdot 600.000) \cdot (400.000 + 2 \cdot 800.000)}, "
                r"2 \cdot 500.000 + 0.69 \cdot (300.000 + 2 \cdot 600.000)\right) = 1969.948 \ mm",
            ),
            ("short", r"r_{cont} = 1969.948 \ mm"),
        ],
    )
    def test_latex_reversed_values(self, representation: str, expected: str) -> None:
        """Test the latex representation of the formula with reversed values."""
        # Example values
        d = 500.0
        c_1 = 400.0
        c_2 = 300.0
        l_h1 = 800.0
        l_h2 = 600.0

        # Object to test
        latex = Form6Dot34And35ContourRadiusRectangular(d=d, c_1=c_1, c_2=c_2, l_h1=l_h1, l_h2=l_h2).latex()

        actual = {
            "complete": latex.complete,
            "short": latex.short,
        }

        assert expected == actual[representation], f"{representation} representation failed."



"""Defines a general steel plotter for cross sections and its characteristics."""

import matplotlib.pyplot as plt
from matplotlib import patches as mplpatches
from matplotlib.patches import Polygon as MplPolygon
from shapely.geometry import Point

from blueprints.structural_sections.steel.steel_cross_sections._steel_cross_section import CombinedSteelCrossSection

# Define color
STEEL_COLOR = (0.683, 0.0, 0.0)


def plot_shapes(
    profile: CombinedSteelCrossSection,
    figsize: tuple[float, float] = (15.0, 8.0),
    title: str = "",
    font_size_title: float = 18.0,
    font_size_legend: float = 10.0,
    show: bool = False,
) -> plt.Figure:
    """
    Plot the given shapes.

    Parameters
    ----------
    profile : SteelCrossSection
        The steel cross-sections to plot.
    figsize : tuple[float, float], optional
        The size of the figure in inches. Default is (15.0, 8.0).
    title : str, optional
        The title of the plot. Default is "".
    font_size_title : float, optional
        The font size of the title. Default is 18.0.
    font_size_legend : float, optional
        The font size of the legend. Default is 10.0.
    show : bool, optional
        Whether to show the plot. Default is False.
    """
    fig, ax = plt.subplots(figsize=figsize)

    for element in profile.elements:
        # Plot the exterior polygon
        x, y = element.cross_section.polygon.exterior.xy
        patch = MplPolygon(xy=list(zip(x, y)), lw=1, fill=True, facecolor=STEEL_COLOR, edgecolor=STEEL_COLOR)
        ax.add_patch(patch)

        # Plot the interior polygons (holes) if any
        for interior in element.cross_section.polygon.interiors:
            x, y = interior.xy
            patch = MplPolygon(xy=list(zip(x, y)), lw=0, fill=True, facecolor="white")
            ax.add_patch(patch)

    # Add dimension lines and centroid
    _add_dimension_lines(ax=ax, profile=profile, centroid=profile.centroid)
    ax.plot(profile.centroid.x, profile.centroid.y, "o", color="black")

    # Add legend text
    profile_section_properties = profile.section_properties(plastic=False, warping=False)
    legend_text = f"""
    {profile.name}\n
    Area: {profile.area:.1f} mm²
    Weight per meter: {profile.weight_per_meter:.1f} kg/m
    Moment of inertia about x: {profile_section_properties.ixx_c:.0f} mm⁴
    Moment of inertia about y: {profile_section_properties.iyy_c:.0f} mm⁴
    """

    # Add the steel quality if all elements have the same material
    if len({element.material.name for element in profile.elements}) == 1:
        legend_text += f"Steel quality: {profile.elements[0].material.name}\n"

    # Get the boundaries of the plot
    _, min_y, _, _ = profile.polygon.bounds
    offset = (profile.width / 2) + (profile.width / 20)

    # Add the legend text to the plot
    ax.annotate(
        xy=(offset, min_y),
        text=legend_text,
        transform=ax.transAxes,
        fontsize=font_size_legend,
    )

    ax.set_xlabel("X")
    ax.set_ylabel("Y")
    ax.set_title(title, fontsize=font_size_title)
    ax.grid(True)
    ax.axis("equal")
    ax.axis("off")
    ax.set_xlim(profile.polygon.bounds[0] - offset, profile.polygon.bounds[2] + offset)

    if show:
        plt.show()  # pragma: no cover

    assert fig is not None

    return fig


def _add_dimension_lines(ax: plt.Axes, profile: CombinedSteelCrossSection, centroid: Point) -> None:
    """Adds dimension lines to show the outer dimensions of the geometry.

    Parameters
    ----------
    ax : plt.Axes
        The matplotlib axes to draw on.
    profile : tuple[CrossSection, ...]
        The cross-sections to plot.
    centroid : Point
        The centroid of the cross-section.
    """
    # Define the offset for the dimension lines
    offset_dimension_lines = max(profile.height, profile.width) / 20
    offset_text = offset_dimension_lines / 2

    # Calculate the bounds of all elements in the geometry
    min_x, min_y, max_x, max_y = profile.polygon.bounds

    # Calculate the width and height of the geometry relative to the centroid
    centroid_width = centroid.x - min_x
    centroid_height = centroid.y - min_y

    # Add the width dimension line (below the geometry)
    diameter_line_style = {
        "arrowstyle": mplpatches.ArrowStyle(stylename="<->", head_length=0.5, head_width=0.5),
    }

    # HORIZONTAL DIMENSION LINES (BELOW THE GEOMETRY)
    # Add the width dimension lines (below the geometry)
    ax.annotate(
        text="",
        xy=(min_x, min_y - offset_dimension_lines * 2),
        xytext=(max_x, min_y - offset_dimension_lines * 2),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        annotation_clip=False,
    )
    ax.text(
        s=f"b= {profile.width:.1f} mm",
        x=(min_x + max_x) / 2,
        y=min_y - offset_dimension_lines * 2 + offset_text,
        verticalalignment="top",
        horizontalalignment="center",
        fontsize=10,
    )

    # Add the height dimension lines (on the left side of the geometry)
    ax.annotate(
        text="",
        xy=(min_x, min_y - offset_dimension_lines),
        xytext=(centroid.x, min_y - offset_dimension_lines),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        annotation_clip=False,
    )
    ax.text(
        s=f"{centroid_width:.1f} mm",
        x=(min_x + centroid.x) / 2,
        y=min_y - offset_dimension_lines + offset_text,
        verticalalignment="top",
        horizontalalignment="center",
        fontsize=10,
    )

    # HEIGHT DIMENSION LINES (ON THE LEFT SIDE OF THE GEOMETRY)
    # Add the distance from the bottom to the centroid (on the left side)
    ax.annotate(
        text="",
        xy=(-max_x - offset_dimension_lines, min_y),
        xytext=(-max_x - offset_dimension_lines, centroid.y),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        rotation=90,
        annotation_clip=False,
    )
    ax.text(
        s=f"{centroid_height:.1f} mm",
        x=-(max_x + offset_dimension_lines + offset_text),
        y=(min_y + centroid.y) / 2,
        verticalalignment="center",
        horizontalalignment="left",
        fontsize=10,
        rotation=90,
    )

    # # Add the height dimension line (on the left side of the geometry)
    ax.annotate(
        text="",
        xy=(-max_x - offset_dimension_lines * 2, max_y),
        xytext=(-max_x - offset_dimension_lines * 2, min_y),
        verticalalignment="center",
        horizontalalignment="center",
        arrowprops=diameter_line_style,
        rotation=90,
        annotation_clip=False,
    )
    ax.text(
        s=f"h= {profile.height:.1f} mm",
        x=-(max_x + offset_dimension_lines * 2 + offset_text),
        y=(min_y + max_y) / 2,
        verticalalignment="center",
        horizontalalignment="left",
        fontsize=10,
        rotation=90,
    )
